#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Author: Andrew Lee, Alejandro Garciadiego
"""
import pytest
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Set,
    value,
    Var,
    units as pyunits,
)
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.unittest import assertStructuredAlmostEqual

from idaes.core import Component
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    fixed_variables_set,
    activated_constraints_set,
)
from idaes.core.solvers import get_solver

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE

from idaes.models.properties.modular_properties.examples.ASU_PR import (
    configuration_Dowling_2015,
)

from idaes.models.properties.tests.test_harness import PropertyTestHarness
from idaes.models.properties.modular_properties.eos.ceos import cubic_roots_available

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


def _as_quantity(x):
    unit = pyunits.get_units(x)
    if unit is None:
        unit = pyunits.dimensionless
    return value(x) * unit._get_pint_unit()


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class TestASUPR(PropertyTestHarness):
    def configure(self):
        self.prop_pack = GenericParameterBlock
        self.param_args = configuration_Dowling_2015
        self.prop_args = {}
        self.has_density_terms = False


# Test for configuration dictionaries with parameters from Properties of Gases
# and liquids 3rd edition and Dowling 2015
# Stream values extracted from:
# A framework for efficient large scale equation-oriented flowsheet optimization (2015)
#     Computers and Chemical Engineering - Alexander W. Dowling
class TestParamBlock(object):
    @pytest.mark.unit
    def test_build(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration_Dowling_2015)

        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 2
        for i in model.params.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.params.Liq.is_liquid_phase()
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 3
        for i in model.params.component_list:
            assert i in ["nitrogen", "argon", "oxygen"]
            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 6
        for i in model.params._phase_component_set:
            assert i in [
                ("Liq", "nitrogen"),
                ("Liq", "argon"),
                ("Liq", "oxygen"),
                ("Vap", "nitrogen"),
                ("Vap", "argon"),
                ("Vap", "oxygen"),
            ]

        assert model.params.config.state_definition == FTPx

        assertStructuredAlmostEqual(
            model.params.config.state_bounds,
            {
                "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
                "temperature": (10, 300, 350, pyunits.K),
                "pressure": (5e4, 1e5, 1e7, pyunits.Pa),
            },
            item_callback=_as_quantity,
        )

        assert model.params.config.phase_equilibrium_state == {
            ("Vap", "Liq"): SmoothVLE
        }

        assert isinstance(model.params.phase_equilibrium_idx, Set)
        assert len(model.params.phase_equilibrium_idx) == 3
        for i in model.params.phase_equilibrium_idx:
            assert i in ["PE1", "PE2", "PE3"]

        assert model.params.phase_equilibrium_list == {
            "PE1": ["nitrogen", ("Vap", "Liq")],
            "PE2": ["argon", ("Vap", "Liq")],
            "PE3": ["oxygen", ("Vap", "Liq")],
        }

        assert model.params.pressure_ref.value == 101325
        assert model.params.temperature_ref.value == 298.15

        assert model.params.nitrogen.mw.value == 28.0135e-3
        assert model.params.nitrogen.pressure_crit.value == 33.943875e5
        assert model.params.nitrogen.temperature_crit.value == 126.2

        assert model.params.argon.mw.value == 39.948e-3
        assert model.params.argon.pressure_crit.value == 48.737325e5
        assert model.params.argon.temperature_crit.value == 150.86

        assert model.params.oxygen.mw.value == 31.999e-3
        assert model.params.oxygen.pressure_crit.value == 50.45985e5
        assert model.params.oxygen.temperature_crit.value == 154.58

        assert_units_consistent(model)


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration_Dowling_2015)

        model.props = model.params.build_state_block([1], defined_state=True)

        return model

    @pytest.mark.unit
    def test_build(self, model):
        # Check state variable values and bounds
        assert isinstance(model.props[1].flow_mol, Var)
        assert value(model.props[1].flow_mol) == 100
        assert model.props[1].flow_mol.ub == 1000
        assert model.props[1].flow_mol.lb == 0

        assert isinstance(model.props[1].pressure, Var)
        assert value(model.props[1].pressure) == 1e5
        assert model.props[1].pressure.ub == 1e7
        assert model.props[1].pressure.lb == 5e4

        assert isinstance(model.props[1].temperature, Var)
        assert value(model.props[1].temperature) == 300
        assert model.props[1].temperature.ub == 350
        assert model.props[1].temperature.lb == 10

        assert isinstance(model.props[1].mole_frac_comp, Var)
        assert len(model.props[1].mole_frac_comp) == 3
        for i in model.props[1].mole_frac_comp:
            assert value(model.props[1].mole_frac_comp[i]) == 1 / 3

        assert_units_consistent(model)

    @pytest.mark.unit
    def test_define_state_vars(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol", "mole_frac_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_port_members(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol", "mole_frac_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_display_vars(self, model):
        sv = model.props[1].define_display_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in [
                "Total Molar Flowrate",
                "Total Mole Fraction",
                "Temperature",
                "Pressure",
            ]

    @pytest.mark.component
    def test_initialize(self, model):
        # Fix state
        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(77.5295)
        model.props[1].pressure.fix(101000)
        model.props[1].mole_frac_comp["nitrogen"].fix(0.9823)
        model.props[1].mole_frac_comp["argon"].fix(0.00803)
        model.props[1].mole_frac_comp["oxygen"].fix(0.00967)

        assert degrees_of_freedom(model.props[1]) == 0

        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={"tol": 1e-6})

        assert degrees_of_freedom(model) == 0

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, model):
        # Check phase equilibrium results
        assert model.props[1].mole_frac_phase_comp[
            "Liq", "nitrogen"
        ].value == pytest.approx(0.9663, abs=1e-3)
        assert model.props[1].mole_frac_phase_comp[
            "Vap", "nitrogen"
        ].value == pytest.approx(0.9890, abs=1e-3)
        assert model.props[1].phase_frac["Vap"].value == pytest.approx(0.703, abs=1e-3)
        assert value(model.props[1].enth_mol_phase["Vap"]) == pytest.approx(
            -6567, abs=1e1
        )

    @pytest.mark.unit
    def test_report(self, model):
        model.props[1].report()

    @pytest.mark.component
    def test_SF0(self, model):
        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(300.00)
        model.props[1].pressure.fix(353140)
        model.props[1].mole_frac_comp["nitrogen"].fix(0.7800)
        model.props[1].mole_frac_comp["argon"].fix(0.0100)
        model.props[1].mole_frac_comp["oxygen"].fix(0.2100)

        assert degrees_of_freedom(model.props[1]) == 0

        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={"tol": 1e-6})

        assert degrees_of_freedom(model) == 0

        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

        assert model.props[1].mole_frac_phase_comp[
            "Vap", "nitrogen"
        ].value == pytest.approx(0.7800, abs=1e-3)
        assert model.props[1].phase_frac["Vap"].value == pytest.approx(1.0, abs=1e-3)
        assert value(model.props[1].enth_mol_phase["Vap"]) == pytest.approx(
            25.7, abs=1e1
        )

    @pytest.mark.component
    def test_SFIL3(self, model):
        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(92.88)
        model.props[1].pressure.fix(353140)
        model.props[1].mole_frac_comp["nitrogen"].fix(0.6653)
        model.props[1].mole_frac_comp["argon"].fix(0.0140)
        model.props[1].mole_frac_comp["oxygen"].fix(0.3207)

        assert degrees_of_freedom(model.props[1]) == 0

        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={"tol": 1e-6})

        assert degrees_of_freedom(model) == 0

        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

        assert model.props[1].mole_frac_phase_comp[
            "Liq", "nitrogen"
        ].value == pytest.approx(0.6653, abs=1e-3)
        assert model.props[1].phase_frac["Vap"].value == pytest.approx(0.0, abs=1e-3)
        assert value(model.props[1].enth_mol_phase["Liq"]) == pytest.approx(
            -11662.4, abs=1e1
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains utilities to provide variable and expression scaling
factors by providing an expression to calculate them via a suffix.

The main purpose of this code is to use the calculate_scaling_factors function
to calculate scaling factors to be used with the Pyomo scaling transformation or
with solvers. A user can provide a scaling_expression suffix to calculate scale
factors from existing variable scaling factors. This allows scaling factors from
a small set of fundamental variables to be propagated to the rest of the model.

The scaling_expression suffix contains Pyomo expressions with model variables.
The expressions can be evaluated with variable scaling factors in place of
variables to calculate additional scaling factors.
"""

__author__ = "John Eslick, Tim Bartholomew, Robert Parker, Andrew Lee"

import math
import sys

import scipy.sparse.linalg as spla
import scipy.linalg as la

import pyomo.environ as pyo
from pyomo.core.base.var import VarData
from pyomo.core.base.param import ParamData
from pyomo.core.expr.visitor import identify_variables
from pyomo.network import Arc
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
from pyomo.contrib.pynumero.asl import AmplInterface
from pyomo.common.modeling import unique_component_name
from pyomo.core.base.constraint import ConstraintData
from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.dae import DerivativeVar
from pyomo.dae.flatten import slice_component_along_sets
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.core import expr as EXPR
from pyomo.common.numeric_types import native_types
from pyomo.core.base.units_container import _PyomoUnit

import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


def __none_left_mult(x, y):
    """PRIVATE FUNCTION, If x is None return None, else return x * y"""
    if x is not None:
        return x * y
    return None


def __scale_constraint(c, v):
    """PRIVATE FUNCTION, scale Constraint c to value v"""
    if c.equality:
        c.set_value((c.lower * v, c.body * v))
    else:
        c.set_value(
            (__none_left_mult(c.lower, v), c.body * v, __none_left_mult(c.upper, v))
        )


def scale_arc_constraints(blk):
    """Find Arc constraints in a block and its subblocks.  Then scale them based
    on the minimum scaling factor of the variables in the constraint.

    Args:
        blk: Block in which to look for Arc constraints to scale.

    Returns:
        None
    """
    for arc in blk.component_data_objects(Arc, descend_into=True):
        arc_block = arc.expanded_block
        if arc_block is None:  # arc not expanded or port empty?
            _log.warning(
                f"{arc} has no constraints. Has the Arc expansion transform "
                "been applied?"
            )
            continue
        warning = (
            "Automatic scaling for arc constraints is supported for "
            "only the Equality rule. Variable {name} on Port {port} was "
            "created with a different rule, so the corresponding constraint "
            "on {arc_name} will not be scaled."
        )
        port1 = arc.ports[0]
        port2 = arc.ports[1]
        for name in port1.vars.keys():
            if not port1.is_equality(name):
                _log.warning(
                    warning.format(name=name, port=port1.name, arc_name=arc.name)
                )
                continue
            if not port2.is_equality(name):
                _log.warning(
                    warning.format(name=name, port=port2.name, arc_name=arc.name)
                )
                continue
            con = getattr(arc_block, name + "_equality")
            for i, c in con.items():
                if i is None:
                    sf = min_scaling_factor([port1.vars[name], port2.vars[name]])
                else:
                    sf = min_scaling_factor([port1.vars[name][i], port2.vars[name][i]])
                constraint_scaling_transform(c, sf)


def map_scaling_factor(components, default=1, warning=False, func=min, hint=None):
    """Map get_scaling_factor to an iterable of Pyomo components, and call func
    on the result.  This could be use, for example, to get the minimum or
    maximum scaling factor of a set of components.

    Args:
        components: Iterable yielding Pyomo components
        default: The default value used when a scaling factor is missing. The
            default is default=1.
        warning: Log a warning for missing scaling factors
        func: The function to call on the resulting iterable of scaling factors.
            The default is min().
        hint: Paired with warning=True, this is a string to indicate where the
            missing scaling factor was being accessed, to easier diagnose issues.

    Returns:
        The result of func on the set of scaling factors
    """
    return func(
        map(
            lambda x: get_scaling_factor(
                x, default=default, warning=warning, hint=hint
            ),
            components,
        )
    )


def min_scaling_factor(components, default=1, warning=True, hint=None):
    """Map get_scaling_factor to an iterable of Pyomo components, and get the
    minimum scaling factor.

    Args:
        iter: Iterable yielding Pyomo components
        default: The default value used when a scaling factor is missing.  If
            None, this will raise an exception when scaling factors are missing.
            The default is default=1.
        warning: Log a warning for missing scaling factors
        hint: Paired with warning=True, this is a string to indicate where the
            missing scaling factor was being accessed, to easier diagnose issues.

    Returns:
        Minimum scaling factor of the components in iter
    """
    return map_scaling_factor(
        components, default=default, warning=warning, func=min, hint=hint
    )


def propagate_indexed_component_scaling_factors(
    blk, typ=None, overwrite=False, descend_into=True
):
    """Use the parent component scaling factor to set all component data object
    scaling factors.

    Args:
        blk: The block on which to search for components
        typ: Component type(s) (default=(Var, Constraint, Expression, Param))
        overwrite: if a data object already has a scaling factor should it be
            overwrittten (default=False)
        descend_into: descend into child blocks (default=True)
    """
    if typ is None:
        typ = (pyo.Var, pyo.Constraint, pyo.Expression)

    for c in blk.component_objects(typ, descend_into=descend_into):
        if get_scaling_factor(c) is not None and c.is_indexed():
            for cdat in c.values():
                if overwrite or get_scaling_factor(cdat) is None:
                    set_scaling_factor(cdat, get_scaling_factor(c))


def calculate_scaling_factors(blk):
    """Look for calculate_scaling_factors methods and run them. This uses a
    recursive function to execute the subblock calculate_scaling_factors
    methods first.
    """

    def cs(blk2):
        """Recursive function for to do subblocks first"""
        for b in blk2.component_data_objects(pyo.Block, descend_into=False):
            cs(b)
        if hasattr(blk2, "calculate_scaling_factors"):
            blk2.calculate_scaling_factors()

    # Call recursive function to run calculate_scaling_factors on blocks from
    # the bottom up.
    cs(blk)
    # If a scale factor is set for an indexed component, propagate it to the
    # component data if a scale factor hasn't already been explicitly set
    propagate_indexed_component_scaling_factors(blk)
    # Use the variable scaling factors to scale the arc constraints.
    scale_arc_constraints(blk)


def set_scaling_factor(c, v, data_objects=True, overwrite=True):
    """Set a scaling factor for a model component. This function creates the
    scaling_factor suffix if needed.

    Args:
        c: component to supply scaling factor for
        v: scaling factor
        data_objects: set scaling factors for indexed data objects (default=True)
        overwrite: whether to overwrite an existing scaling factor
    Returns:
        None
    """
    if isinstance(c, (float, int)):
        # property packages can return 0 for material balance terms on components
        # doesn't exist.  This handles the case where you get a constant 0 and
        # need its scale factor to scale the mass balance.
        return 1
    try:
        suf = c.parent_block().scaling_factor

    except AttributeError:
        c.parent_block().scaling_factor = pyo.Suffix(direction=pyo.Suffix.EXPORT)
        suf = c.parent_block().scaling_factor

    if not overwrite:
        try:
            tmp = suf[c]  # pylint: disable=unused-variable
            # Able to access suffix value for c, so return without setting scaling factor
            return
        except KeyError:
            # No value exists yet for c, go ahead and set it
            pass
    suf[c] = v
    if data_objects and c.is_indexed():
        for cdat in c.values():
            if not overwrite:
                try:
                    tmp = suf[cdat]
                    continue
                except KeyError:
                    pass
            suf[cdat] = v


def get_scaling_factor(c, default=None, warning=False, exception=False, hint=None):
    """Get a component scale factor.

    Args:
        c: component
        default: value to return if no scale factor exists (default=None)
        warning: whether to log a warning if a scaling factor is not found
                 (default=False)
        exception: whether to raise an Exception if a scaling factor is not
                   found (default=False)
        hint: (str) a string to add to the warning or exception message to help
            locate the source.

    Returns:
        scaling factor (float)
    """
    try:
        sf = c.parent_block().scaling_factor[c]
    except (AttributeError, KeyError):
        if not isinstance(c, (pyo.Param, ParamData)):
            if hint is None:
                h = ""
            else:
                h = f", {hint}"
            if warning:
                if hasattr(c, "is_component_type") and c.is_component_type():
                    _log.warning(f"Missing scaling factor for {c}{h}")
                else:
                    _log.warning(f"Trying to get scaling factor for unnamed expr {h}")
            if exception and default is None:
                if hasattr(c, "is_component_type") and c.is_component_type():
                    _log.error(f"Missing scaling factor for {c}{h}")
                else:
                    _log.error(f"Trying to get scaling factor for unnamed expr {h}")
                raise
            sf = default
        else:
            # Params can just use current value (as long it is not 0)
            val = pyo.value(c)
            if not val == 0:
                sf = abs(1 / pyo.value(c))
            else:
                sf = 1
    return sf


def set_and_get_scaling_factor(c, default, warning=False, exception=False, hint=None):
    """Checks whether a scaling factor exists for a component, sets the scaling factor
    if it doesn't exist, then returns the scaling factor on the component (which is
    the default value if it wasn't set originally).

    Args:
        c: component
        default: default value to use for scaling factor of c if there is none
        warning: whether to log a warning if a scaling factor is not found
                 (default=False)
        exception: whether to raise an Exception if a scaling factor is not
                   found (default=False)
        hint: (str) a string to add to the warning or exception message to help
            locate the source.

    Returns:
        scaling factor (float)
    """
    if c.is_indexed():
        raise AttributeError(
            f"Ambiguous which scaling factor to return for indexed component {c.name}."
        )
    sf = get_scaling_factor(c, warning=warning, exception=exception, hint=hint)
    if sf is None:
        sf = default
        set_scaling_factor(c, sf, data_objects=False)
    return sf


def unset_scaling_factor(c, data_objects=True):
    """Delete a component scaling factor.

    Args:
        c: component

    Returns:
        None
    """
    try:
        del c.parent_block().scaling_factor[c]
    except (AttributeError, KeyError):
        pass  # no scaling factor suffix, is fine
    try:
        if data_objects and c.is_indexed():
            for cdat in c.values():
                del cdat.parent_block().scaling_factor[cdat]
    except (AttributeError, KeyError):
        pass  # no scaling factor suffix, is fine


def populate_default_scaling_factors(c):
    """
    Method to set default scaling factors for a number of common quantities
    based of typical values expressed in SI units. Values are converted to
    those used by the property package using Pyomo's unit conversion tools.
    """
    units = c.get_metadata().derived_units

    si_scale = {
        "temperature": (100 * pyo.units.K, "temperature"),
        "pressure": (1e5 * pyo.units.Pa, "pressure"),
        "dens_mol_phase": (100 * pyo.units.mol / pyo.units.m**3, "density_mole"),
        "enth_mol": (1e4 * pyo.units.J / pyo.units.mol, "energy_mole"),
        "entr_mol": (100 * pyo.units.J / pyo.units.mol / pyo.units.K, "entropy_mole"),
        "fug_phase_comp": (1e4 * pyo.units.Pa, "pressure"),
        "fug_coeff_phase_comp": (1 * pyo.units.dimensionless, None),
        "gibbs_mol": (1e4 * pyo.units.J / pyo.units.mol, "energy_mole"),
        "mole_frac_comp": (0.001 * pyo.units.dimensionless, None),
        "mole_frac_phase_comp": (0.001 * pyo.units.dimensionless, None),
        "mw": (1e-3 * pyo.units.kg / pyo.units.mol, "molecular_weight"),
        "mw_comp": (1e-3 * pyo.units.kg / pyo.units.mol, "molecular_weight"),
        "mw_phase": (1e-3 * pyo.units.kg / pyo.units.mol, "molecular_weight"),
    }

    for p, f in si_scale.items():
        # If a default scaling factor exists, do not over write it
        if p not in c.default_scaling_factor.keys():
            if f[1] is not None:
                v = pyo.units.convert(f[0], to_units=units[f[1]])
            else:
                v = f[0]

            sf = 1 / (10 ** round(math.log10(pyo.value(v))))

            c.set_default_scaling(p, sf)


def __set_constraint_transform_applied_scaling_factor(c, v):
    """PRIVATE FUNCTION Set the scaling factor used to transform a constraint.
    This is used to keep track of scaling transformations that have been applied
    to constraints.

    Args:
        c: component to supply scaling factor for
        v: scaling factor
    Returns:
        None
    """
    try:
        c.parent_block().constraint_transformed_scaling_factor[c] = v
    except AttributeError:
        c.parent_block().constraint_transformed_scaling_factor = pyo.Suffix(
            direction=pyo.Suffix.LOCAL
        )
        c.parent_block().constraint_transformed_scaling_factor[c] = v


def get_constraint_transform_applied_scaling_factor(c, default=None):
    """Get a the scale factor that was used to transform a
    constraint.

    Args:
        c: constraint data object
        default: value to return if no scaling factor exists (default=None)

    Returns:
        The scaling factor that has been used to transform the constraint or the
        default.
    """
    try:
        sf = c.parent_block().constraint_transformed_scaling_factor.get(c, default)
    except AttributeError:
        sf = default  # when there is no suffix
    return sf


def __unset_constraint_transform_applied_scaling_factor(c):
    """PRIVATE FUNCTION: Delete the recorded scale factor that has been used
    to transform constraint c.  This is used when undoing a constraint
    transformation.
    """
    try:
        del c.parent_block().constraint_transformed_scaling_factor[c]
    except AttributeError:
        pass  # no scaling factor suffix, is fine
    except KeyError:
        pass  # no scaling factor is fine


def constraint_scaling_transform(c, s, overwrite=True):
    """This transforms a constraint by the argument s.  The scaling factor
    applies to original constraint (e.g. if one where to call this twice in a row
    for a constraint with a scaling factor of 2, the original constraint would
    still, only be scaled by a factor of 2.)

    Args:
        c: Pyomo constraint
        s: scale factor applied to the constraint as originally written
        overwrite: overwrite existing scaling factors if present (default=True)

    Returns:
        None
    """
    # Want to clear away any units that may have incidentally become attached to s
    s = pyo.value(s)
    if not isinstance(c, ConstraintData):
        raise TypeError(f"{c} is not a constraint or is an indexed constraint")
    st = get_constraint_transform_applied_scaling_factor(c, default=None)

    if not overwrite and st is not None:
        # Existing scaling factor and overwrite False, do nothing
        return

    if st is None:
        # If no existing scaling factor, use value of 1
        st = 1

    v = s / st
    __scale_constraint(c, v)
    __set_constraint_transform_applied_scaling_factor(c, s)


def constraint_scaling_transform_undo(c):
    """The undoes the scaling transforms previously applied to a constraint.

    Args:
        c: Pyomo constraint

    Returns:
        None
    """
    if not isinstance(c, ConstraintData):
        raise TypeError(f"{c} is not a constraint or is an indexed constraint")
    v = get_constraint_transform_applied_scaling_factor(c)
    if v is None:
        return  # hasn't been transformed, so nothing to do.
    __scale_constraint(c, 1 / v)
    __unset_constraint_transform_applied_scaling_factor(c)


def unscaled_variables_generator(blk, descend_into=True, include_fixed=False):
    """Generator for unscaled variables

    Args:
        block

    Yields:
        variables with no scale factor
    """
    for v in blk.component_data_objects(pyo.Var, descend_into=descend_into):
        if v.fixed and not include_fixed:
            continue
        if get_scaling_factor(v) is None:
            yield v


def list_unscaled_variables(
    blk: pyo.Block, descend_into: bool = True, include_fixed: bool = False
):
    """
    Return a list of variables which do not have a scaling factor assigned
    Args:
        blk: block to check for unscaled variables
        descend_into: bool indicating whether to check variables in sub-blocks
        include_fixed: bool indicating whether to include fixed Vars in list

    Returns:
        list of unscaled variable data objects
    """
    return [c for c in unscaled_variables_generator(blk, descend_into, include_fixed)]


def unscaled_constraints_generator(blk, descend_into=True):
    """Generator for unscaled constraints

    Args:
        block

    Yields:
        constraints with no scale factor
    """
    for c in blk.component_data_objects(
        pyo.Constraint, active=True, descend_into=descend_into
    ):
        if (
            get_scaling_factor(c) is None
            and get_constraint_transform_applied_scaling_factor(c) is None
        ):
            yield c


def list_unscaled_constraints(blk: pyo.Block, descend_into: bool = True):
    """
    Return a list of constraints which do not have a scaling factor assigned
    Args:
        blk: block to check for unscaled constraints
        descend_into: bool indicating whether to check constraints in sub-blocks

    Returns:
        list of unscaled constraint data objects
    """
    return [c for c in unscaled_constraints_generator(blk, descend_into)]


def constraints_with_scale_factor_generator(blk, descend_into=True):
    """Generator for constraints scaled by a scaling factor, may or not have
    been transformed.

    Args:
        block

    Yields:
        constraint with a scale factor, scale factor
    """
    for c in blk.component_data_objects(
        pyo.Constraint, active=True, descend_into=descend_into
    ):
        s = get_scaling_factor(c)
        if s is not None:
            yield c, s


def badly_scaled_var_generator(
    blk, large=1e4, small=1e-3, zero=1e-10, descend_into=True, include_fixed=False
):
    """This provides a rough check for variables with poor scaling based on
    their current scale factors and values. For each potentially poorly scaled
    variable it returns the var and its current scaled value.

    Note that while this method is a reasonable heuristic for non-negative
    variables like (absolute) temperature and pressure, molar flows, etc., it
    can be misleading for variables like enthalpies and fluxes.

    Args:
        blk: pyomo block
        large: Magnitude that is considered to be too large
        small: Magnitude that is considered to be too small
        zero: Magnitude that is considered to be zero, variables with a value of
            zero are okay, and not reported.

    Yields:
        variable data object, current absolute value of scaled value
    """
    for v in blk.component_data_objects(pyo.Var, descend_into=descend_into):
        if v.fixed and not include_fixed:
            continue
        val = pyo.value(v, exception=False)
        if val is None:
            continue
        sf = get_scaling_factor(v, default=1)
        sv = abs(val * sf)  # scaled value
        if sv > large:
            yield v, sv
        elif sv < zero:
            continue
        elif sv < small:
            yield v, sv


def list_badly_scaled_variables(
    blk,
    large: float = 1e4,
    small: float = 1e-3,
    zero: float = 1e-10,
    descend_into: bool = True,
    include_fixed: bool = False,
):
    """Return a list of variables with poor scaling based on
    their current scale factors and values. For each potentially poorly scaled
    variable it returns the var and its current scaled value.

    Note that while this method is a reasonable heuristic for non-negative
    variables like (absolute) temperature and pressure, molar flows, etc., it
    can be misleading for variables like enthalpies and fluxes.

    Args:
        blk: pyomo block
        large: Magnitude that is considered to be too large
        small: Magnitude that is considered to be too small
        zero: Magnitude that is considered to be zero, variables with a value of
            zero are okay, and not reported.
        descend_into: bool indicating whether to check constraints in sub-blocks
        include_fixed: bool indicating whether to include fixed Vars in list


    Returns:
        list of tuples containing (variable data object, current absolute value of scaled value)
    """
    return [
        c
        for c in badly_scaled_var_generator(
            blk, large, small, zero, descend_into, include_fixed
        )
    ]


def constraint_autoscale_large_jac(
    m,
    ignore_constraint_scaling=False,
    ignore_variable_scaling=False,
    max_grad=100,
    min_scale=1e-6,
    no_scale=False,
    equality_constraints_only=False,
):
    """Automatically scale constraints based on the Jacobian.  This function
    imitates Ipopt's default constraint scaling.  This scales constraints down
    to avoid extremely large values in the Jacobian.  This function also returns
    the unscaled and scaled Jacobian matrixes and the Pynumero NLP which can be
    used to identify the constraints and variables corresponding to the rows and
    comlumns.

    Args:
        m: model to scale
        ignore_constraint_scaling: ignore existing constraint scaling
        ignore_variable_scaling: ignore existing variable scaling
        max_grad: maximum value in Jacobian after scaling, subject to minimum
            scaling factor restriction.
        min_scale: minimum scaling factor allowed, keeps constraints from being
            scaled too much.
        no_scale: just calculate the Jacobian and scaled Jacobian, don't scale
            anything
        equality_constraints_only: Include only the equality constraints in the
            Jacobian

    Returns:
        unscaled Jacobian CSR from, scaled Jacobian CSR from, Pynumero NLP
    """
    # Pynumero requires an objective, but I don't, so let's see if we have one
    n_obj = 0
    for c in m.component_data_objects(pyo.Objective, active=True):
        n_obj += 1
    # Add an objective if there isn't one
    if n_obj == 0:
        dummy_objective_name = unique_component_name(m, "objective")
        setattr(m, dummy_objective_name, pyo.Objective(expr=0))
    # Create NLP and calculate the objective
    if not AmplInterface.available():
        raise RuntimeError("Pynumero not available.")
    nlp = PyomoNLP(m)
    if equality_constraints_only:
        jac = nlp.evaluate_jacobian_eq().tocsr()
    else:
        jac = nlp.evaluate_jacobian().tocsr()
    # Get lists of variables and constraints to translate Jacobian indexes
    # save them on the NLP for later, since generating them seems to take a while
    if equality_constraints_only:
        nlp.clist = clist = nlp.get_pyomo_equality_constraints()
    else:
        nlp.clist = clist = nlp.get_pyomo_constraints()
    nlp.vlist = vlist = nlp.get_pyomo_variables()
    # Create a scaled Jacobian to account for variable scaling, for now ignore
    # constraint scaling
    jac_scaled = jac.copy()
    for i, c in enumerate(clist):
        for j in jac_scaled[i].indices:
            v = vlist[j]
            if ignore_variable_scaling:
                sv = 1
            else:
                sv = get_scaling_factor(v, default=1)
            jac_scaled[i, j] = jac_scaled[i, j] / sv
    # calculate constraint scale factors
    for i, c in enumerate(clist):
        sc = get_scaling_factor(c, default=1)
        if not no_scale:
            if ignore_constraint_scaling or get_scaling_factor(c) is None:
                sc = 1
                row = jac_scaled[i]
                for d in row.indices:
                    row[0, d] = abs(row[0, d])
                mg = row.max()
                if mg > max_grad:
                    sc = max(min_scale, max_grad / mg)
                set_scaling_factor(c, sc)
        for j in jac_scaled[i].indices:
            # update the scaled jacobian
            jac_scaled[i, j] = jac_scaled[i, j] * sc
    # delete dummy objective
    if n_obj == 0:
        delattr(m, dummy_objective_name)
    return jac, jac_scaled, nlp


def get_jacobian(m, scaled=True, equality_constraints_only=False):
    """
    Get the Jacobian matrix at the current model values. This function also
    returns the Pynumero NLP which can be used to identify the constraints and
    variables corresponding to the rows and columns.

    Args:
        m: model to get Jacobian from
        scaled: if True return scaled Jacobian, else get unscaled
        equality_constraints_only: Only include equality constraints in the
            Jacobian calculated and scaled

    Returns:
        Jacobian matrix in Scipy CSR format, Pynumero nlp
    """
    jac, jac_scaled, nlp = constraint_autoscale_large_jac(
        m, no_scale=True, equality_constraints_only=equality_constraints_only
    )
    if scaled:
        return jac_scaled, nlp
    else:
        return jac, nlp


def extreme_jacobian_entries(
    m=None, scaled=True, large=1e4, small=1e-4, zero=1e-10, jac=None, nlp=None
):
    """
    Show very large and very small Jacobian entries.

    Args:
        m: model
        scaled: if true use scaled Jacobian
        large: >= to this value is considered large
        small: <= to this and >= zero is considered small

    Returns:
        (list of tuples), Jacobian entry, Constraint, Variable
    """
    if jac is None or nlp is None:
        jac, nlp = get_jacobian(m, scaled)
    el = []
    for i, c in enumerate(nlp.clist):
        for j in jac[i].indices:
            v = nlp.vlist[j]
            e = abs(jac[i, j])
            if (e <= small and e > zero) or e >= large:
                el.append((e, c, v))
    return el


def extreme_jacobian_rows(
    m=None, scaled=True, large=1e4, small=1e-4, jac=None, nlp=None
):
    """
    Show very large and very small Jacobian rows. Typically indicates a badly-
    scaled constraint.

    Args:
        m: model
        scaled: if true use scaled Jacobian
        large: >= to this value is considered large
        small: <= to this is considered small

    Returns:
        (list of tuples), Row norm, Constraint
    """
    # Need both jac for the linear algebra and nlp for constraint names
    if jac is None or nlp is None:
        jac, nlp = get_jacobian(m, scaled)
    el = []
    for i, c in enumerate(nlp.clist):
        norm = 0
        # Calculate L2 norm
        for j in jac[i].indices:
            norm += jac[i, j] ** 2
        norm = norm**0.5
        if norm <= small or norm >= large:
            el.append((norm, c))
    return el


def extreme_jacobian_columns(
    m=None, scaled=True, large=1e4, small=1e-4, jac=None, nlp=None
):
    """
    Show very large and very small Jacobian columns. A more reliable indicator
    of a badly-scaled variable than badly_scaled_var_generator.

    Args:
        m: model
        scaled: if true use scaled Jacobian
        large: >= to this value is considered large
        small: <= to this is considered small

    Returns:
        (list of tuples), Column norm, Variable
    """
    # Need both jac for the linear algebra and nlp for variable names
    if jac is None or nlp is None:
        jac, nlp = get_jacobian(m, scaled)
    jac = jac.tocsc()
    el = []
    for j, v in enumerate(nlp.vlist):
        norm = 0
        # Calculate L2 norm
        for i in jac.getcol(j).indices:
            norm += jac[i, j] ** 2
        norm = norm**0.5
        if norm <= small or norm >= large:
            el.append((norm, v))
    return el


def jacobian_cond(m=None, scaled=True, order=None, pinv=False, jac=None):
    """
    Get the condition number of the scaled or unscaled Jacobian matrix of a model.

    Args:
        m: calculate the condition number of the Jacobian from this model.
        scaled: if True use scaled Jacobian, else use unscaled
        order: norm order, None = Frobenius, see scipy.sparse.linalg.norm for more
        pinv: Use pseudoinverse, works for non-square matrices
        jac: (optional) previously calculated Jacobian

    Returns:
        (float) Condition number
    """
    if jac is None:
        jac, _ = get_jacobian(m, scaled)
    jac = jac.tocsc()
    if jac.shape[0] != jac.shape[1] and not pinv:
        _log.warning("Nonsquare Jacobian using pseudo inverse")
        pinv = True
    if not pinv:
        jac_inv = spla.inv(jac)
        return spla.norm(jac, order) * spla.norm(jac_inv, order)
    else:
        jac_inv = la.pinv(jac.toarray())
        return spla.norm(jac, order) * la.norm(jac_inv, order)


def scale_time_discretization_equations(blk, time_set, time_scaling_factor):
    """
    Scales time discretization equations generated via a Pyomo discretization
    transformation. Also scales continuity equations for collocation methods
    of discretization that require them.

    Args:
        blk: Block whose time discretization equations are being scaled
        time_set: Time set object. For an IDAES flowsheet object fs, this is fs.time.
        time_scaling_factor: Scaling factor to use for time

    Returns:
        None
    """

    tname = time_set.local_name

    # Copy and pasted from solvers.petsc.find_discretization_equations then modified
    for var in blk.component_objects(pyo.Var):
        if isinstance(var, DerivativeVar):
            cont_set_set = ComponentSet(var.get_continuousset_list())
            if time_set in cont_set_set:
                if len(cont_set_set) > 1:
                    _log.warning(
                        "IDAES presently does not support automatically scaling discretization equations for "
                        f"second order or higher derivatives like {var.name} that are differentiated at least once with "
                        "respect to time. Please scale the corresponding discretization equation yourself."
                    )
                    continue
                state_var = var.get_state_var()
                parent_block = var.parent_block()

                disc_eq = getattr(parent_block, var.local_name + "_disc_eq")
                # Look for continuity equation, which exists only for collocation with certain sets of polynomials
                try:
                    cont_eq = getattr(
                        parent_block, state_var.local_name + "_" + tname + "_cont_eq"
                    )
                except AttributeError:
                    cont_eq = None

                deriv_dict = dict(
                    (key, pyo.Reference(slc))
                    for key, slc in slice_component_along_sets(var, (time_set,))
                )
                state_dict = dict(
                    (key, pyo.Reference(slc))
                    for key, slc in slice_component_along_sets(state_var, (time_set,))
                )
                disc_dict = dict(
                    (key, pyo.Reference(slc))
                    for key, slc in slice_component_along_sets(disc_eq, (time_set,))
                )
                if cont_eq is not None:
                    cont_dict = dict(
                        (key, pyo.Reference(slc))
                        for key, slc in slice_component_along_sets(cont_eq, (time_set,))
                    )
                for key, deriv in deriv_dict.items():
                    state = state_dict[key]
                    disc = disc_dict[key]
                    if cont_eq is not None:
                        cont = cont_dict[key]
                    for t in time_set:
                        s_state = get_scaling_factor(state[t], default=1, warning=True)
                        set_scaling_factor(
                            deriv[t], s_state / time_scaling_factor, overwrite=False
                        )
                        s_deriv = get_scaling_factor(deriv[t])
                        # Check time index to decide what constraints to scale
                        if cont_eq is None:
                            if t == time_set.first() or t == time_set.last():
                                try:
                                    constraint_scaling_transform(
                                        disc[t], s_deriv, overwrite=False
                                    )
                                except KeyError:
                                    # Discretization and continuity equations may or may not exist at the first or last time
                                    # points depending on the method. Backwards skips first, forwards skips last, central skips
                                    # both (which means the user needs to provide additional equations)
                                    pass
                            else:
                                constraint_scaling_transform(
                                    disc[t], s_deriv, overwrite=False
                                )
                        else:
                            # Lagrange-Legendre is a pain, because it has continuity equations on the edges of finite
                            # instead of discretization equations, but no intermediate continuity equations, so we have
                            # to look for both at every timepoint
                            try:
                                constraint_scaling_transform(
                                    disc[t], s_deriv, overwrite=False
                                )
                            except KeyError:
                                if t != time_set.first():
                                    constraint_scaling_transform(
                                        cont[t], s_state, overwrite=False
                                    )


class CacheVars(object):
    """
    A class for saving the values of variables then reloading them,
    usually after they have been used to perform some solve or calculation.
    """

    def __init__(self, vardata_list):
        self.vars = vardata_list
        self.cache = [None for var in self.vars]

    def __enter__(self):
        for i, var in enumerate(self.vars):
            self.cache[i] = var.value
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        for i, var in enumerate(self.vars):
            var.set_value(self.cache[i])


class FlattenedScalingAssignment(object):
    """
    A class to assist in the calculation of scaling factors when a
    variable-constraint assignment can be constructed, especially when
    the variables and constraints are all indexed by some common set(s).
    """

    def __init__(self, scaling_factor, varconlist=None, nominal_index=()):
        """
        Args:
            scaling_factor: A Pyomo scaling_factor Suffix that will hold all
                            the scaling factors calculated
            varconlist: A list of variable, constraint tuples. These variables
                        and constraints should be indexed by the same sets,
                        so they may need to be references-to-slices along some
                        common sets.
            nominal_index: The index of variables and constraints to access
                           when a calculation needs to be performed using
                           data objects.
        """
        if varconlist is None:
            varconlist = []

        self.scaling_factor = scaling_factor
        self.nominal_index = nominal_index
        if nominal_index is None or nominal_index == ():
            self.dim = 0
        else:
            try:
                self.dim = len(nominal_index)
            except TypeError:
                self.dim = 1

        varlist = []
        conlist = []
        for var, con in varconlist:
            varlist.append(var)
            conlist.append(con)
        self.varlist = varlist
        self.conlist = conlist

        data_getter = self.get_representative_data_object
        var_con_data_list = [
            (data_getter(var), data_getter(con)) for var, con in varconlist
        ]
        con_var_data_list = [
            (data_getter(con), data_getter(var)) for var, con in varconlist
        ]
        self.var2con = ComponentMap(var_con_data_list)
        self.con2var = ComponentMap(con_var_data_list)

    def get_representative_data_object(self, obj):
        """
        Gets a data object from an object of the appropriate dimension
        """
        if self.dim == 0:
            # In this way, obj can be a data object and this class can be
            # used even if the assignment is not between "flattened components"
            return obj
        else:
            nominal_index = self.nominal_index
            return obj[nominal_index]

    def calculate_variable_scaling_factor(self, var, include_fixed=False):
        """
        Calculates the scaling factor of a variable based on the
        constraint assigned to it. Loads each variable in that constraint
        with its nominal value (inverse of scaling factor), calculates
        the value of the target variable from the constraint, then sets
        its scaling factor to the inverse of the calculated value.
        """
        vardata = self.get_representative_data_object(var)
        condata = self.var2con[vardata]
        scaling_factor = self.scaling_factor

        in_constraint = list(
            identify_variables(
                condata.expr,
                include_fixed=include_fixed,
            )
        )
        source_vars = [v for v in in_constraint if v is not vardata]
        nominal_source = [1 / scaling_factor[var] for var in source_vars]

        with CacheVars(in_constraint):
            for v, nom_val in zip(source_vars, nominal_source):
                v.set_value(nom_val)
            # This assumes that target var is initialized to a somewhat
            # reasonable value
            calculate_variable_from_constraint(vardata, condata)
            nominal_target = vardata.value
        if nominal_target == 0:
            target_factor = 1.0
        else:
            target_factor = abs(1 / nominal_target)

        if self.dim == 0:
            scaling_factor[var] = target_factor
        else:
            for v in var.values():
                scaling_factor[v] = target_factor

    def set_constraint_scaling_factor(self, con):
        """
        Sets the scaling factor of a constraint to that of its assigned variable
        """
        condata = self.get_representative_data_object(con)
        vardata = self.con2var[condata]
        scaling_factor = self.scaling_factor

        var_factor = scaling_factor[vardata]
        if self.dim == 0:
            scaling_factor[con] = var_factor
        else:
            for c in con.values():
                scaling_factor[c] = var_factor

    def set_derivative_factor_from_state(self, deriv, nominal_wrt=1.0):
        """
        Sets the scaling factor for a DerivativeVar equal to the factor for
        its state var at every index. This method needs access to the
        get_state_var method, so deriv must be an actual DerivativeVar,
        not a reference-to-slice.
        """
        scaling_factor = self.scaling_factor
        state_var = deriv.get_state_var()
        for index, dv in deriv.items():
            state_data = state_var[index]
            nominal_state = 1 / scaling_factor[state_data]
            nominal_deriv = nominal_state / nominal_wrt
            scaling_factor[dv] = 1 / nominal_deriv


# New functions
def set_scaling_from_default(
    component,
    missing: float = None,
    overwrite: bool = False,
    descend_into: bool = True,
    components_to_scale: "List of Pyomo component types" = None,
):
    """
    Set scaling factor(s) for given component from default scaling factor dictionary associated with the parent model.

    This function accepts any type of Pyomo component as an input, and will attempt to apply scaling factors to all
    attached types of components listed in 'components_to_scale' argument. A warning will be logged for any
    component which does not have a default scaling factor assigned.

    Args:
        component: Pyomo component to apply scaling factors to.
        missing: value to use if a component does not have a default scaling factor assigned (default=None).
        overwrite: bool indicating whether to overwrite existing scaling factors (default=False).
        descend_into: bool indicating whether to descend into child Blocks if component is a Block (default=True).
        components_to_scale: list of Pyomo component types to apply scaling factors to if component is a Block (default=[Var]).

    Returns:
        None

    """
    if components_to_scale is None:
        components_to_scale = [pyo.Var]

    if isinstance(component, pyo.Block):
        for c in component.component_data_objects(
            components_to_scale, descend_into=descend_into
        ):
            set_scaling_from_default(c, missing=missing, overwrite=overwrite)
    elif component.is_indexed():
        for k in component.values():
            set_scaling_from_default(k, missing=missing, overwrite=overwrite)
    else:
        if not overwrite:
            # If there is already a scaling factor, we can end here
            sf = get_scaling_factor(
                component, default=None, warning=False, exception=False
            )
            if sf is not None:
                return

        parent = component.parent_block()
        try:
            dsf = parent.get_default_scaling(
                component.parent_component().local_name, index=component.index()
            )
        except AttributeError:
            dsf = None
            _log.warning(
                f"{component.name} block missing 'get_default_scaling()' method, no scaling factor assigned."
            )

        if dsf is not None:
            set_scaling_factor(component, dsf, overwrite=overwrite)
        elif missing is not None:
            _log.warning(
                f"No default scaling factor found for {component.name}, assigning value of {missing} instead."
            )
            set_scaling_factor(component, missing, overwrite=overwrite)
        else:
            _log.warning(
                f"No default scaling factor found for {component.name}, no scaling factor assigned."
            )


def set_variable_scaling_from_current_value(
    component, descend_into: bool = True, overwrite: bool = False
):
    """
    Set scaling factor for variables based on current value. Component argument can be either a Pyomo Var or Block.
    In case of a Block, this function will attempt to scale all variables in the block using their current value,

    Args:
        component: component to scale
        overwrite: bool indicating whether to overwrite existing scaling factors (default=False).
        descend_into: bool indicating whether to descend into child Blocks if component is a Block (default=True).

    Returns:
        None

    """
    if isinstance(component, pyo.Block):
        for c in component.component_data_objects(pyo.Var, descend_into=descend_into):
            set_variable_scaling_from_current_value(c, overwrite=overwrite)
    elif component.is_indexed():
        for k in component.values():
            set_variable_scaling_from_current_value(k, overwrite=overwrite)
    elif not isinstance(component, VarData):
        raise TypeError(
            f"Invalid component type {component.name} (type:{type(component)}). "
            "component argument to set_variable_scaling_from_current_value must be "
            "either a Pyomo Var or Block."
        )
    else:
        if not overwrite:
            # If there is already a scaling factor, we can end here
            sf = get_scaling_factor(
                component, default=None, warning=False, exception=False
            )
            if sf is not None:
                return

        try:
            val = pyo.value(component)

            if val == 0:
                _log.warning(
                    f"Component {component.name} currently has a value of 0; no scaling factor assigned."
                )
            else:
                set_scaling_factor(component, 1 / val, overwrite=overwrite)
        except ValueError:
            _log.warning(
                f"Component {component.name} does not have a current value; no scaling factor assigned."
            )


# TODO: Deprecate in favor of new walker
class NominalValueExtractionVisitor(EXPR.StreamBasedExpressionVisitor):
    """
    Expression walker for collecting scaling factors in an expression and determining the
    expected value of the expression using the scaling factors as nominal inputs.

    Returns a list of expected values for each additive term in the expression.

    In order to properly assess the expected value of terms within functions, the sign
    of each term is maintained throughout thus returned values may be negative. Functions
    using this walker should handle these appropriately.
    """

    def __init__(self, warning: bool = True):
        """
        Visitor class used to determine nominal values of all terms in an expression based on
        scaling factors assigned to the associated variables. Do not use this class directly.

        Args:
            warning: bool indicating whether to log a warning when a
                missing scaling factors is encountered (default=True)

        Notes
        -----
        This class inherits from the :class:`StreamBasedExpressionVisitor` to implement
        a walker that returns the nominal value corresponding to all additive terms in an
        expression.
        There are class attributes (dicts) that map the expression node type to the
        particular method that should be called to return the nominal value of the node based
        on the nominal value of its child arguments. This map is used in exitNode.
        """
        super().__init__()

        self.warning = warning

    def _get_magnitude_base_type(self, node):
        # Get scaling factor for node
        sf = get_scaling_factor(node, default=1, warning=self.warning)

        # Try to determine expected sign of node
        if isinstance(node, pyo.Var):
            ub = node.ub
            lb = node.lb
            domain = node.domain

            # To avoid NoneType errors, assign dummy values in place of None
            if ub is None:
                # No upper bound, take a positive value
                ub = 1000
            if lb is None:
                # No lower bound, take a negative value
                lb = -1000

            if lb >= 0 or domain in [
                pyo.NonNegativeReals,
                pyo.PositiveReals,
                pyo.PositiveIntegers,
                pyo.NonNegativeIntegers,
                pyo.Boolean,
                pyo.Binary,
            ]:
                # Strictly positive
                sign = 1
            elif ub <= 0 or domain in [
                pyo.NegativeReals,
                pyo.NonPositiveReals,
                pyo.NegativeIntegers,
                pyo.NonPositiveIntegers,
            ]:
                # Strictly negative
                sign = -1
            else:
                # Unbounded, see if there is a current value
                try:
                    value = pyo.value(node)
                except ValueError:
                    value = None

                if value is not None and value < 0:
                    # Assigned negative value, assume value will remain negative
                    sign = -1
                else:
                    # Either a positive value or no value, assume positive
                    sign = 1
        elif isinstance(node, pyo.Param):
            domain = node.domain

            if domain in [
                pyo.NonNegativeReals,
                pyo.PositiveReals,
                pyo.PositiveIntegers,
                pyo.NonNegativeIntegers,
                pyo.Boolean,
                pyo.Binary,
            ]:
                # Strictly positive
                sign = 1
            elif domain in [
                pyo.NegativeReals,
                pyo.NonPositiveReals,
                pyo.NegativeIntegers,
                pyo.NonPositiveIntegers,
            ]:
                # Strictly negative
                sign = -1
            else:
                # Unbounded, see if there is a current value
                try:
                    value = pyo.value(node)
                except ValueError:
                    value = None

                if value is not None and value < 0:
                    # Assigned negative value, assume value will remain negative
                    sign = -1
                else:
                    # Either a positive value or no value, assume positive
                    sign = 1
        else:
            # No idea, assume positive
            sign = 1

        try:
            return [sign / sf]
        except ZeroDivisionError:
            raise ValueError(
                f"Found component {node.name} with scaling factor of 0. "
                "Scaling factors should not be set to 0 as this results in "
                "numerical failures."
            )

    def _get_nominal_value_for_sum_subexpression(self, child_nominal_values):
        return sum(i for i in child_nominal_values)

    def _get_nominal_value_for_sum(self, node, child_nominal_values):
        # For sums, collect all child values into a list
        sf = []
        for i in child_nominal_values:
            for j in i:
                sf.append(j)
        return sf

    def _get_nominal_value_for_product(self, node, child_nominal_values):
        assert len(child_nominal_values) == 2

        mag = []
        for i in child_nominal_values[0]:
            for j in child_nominal_values[1]:
                mag.append(i * j)
        return mag

    def _get_nominal_value_for_division(self, node, child_nominal_values):
        assert len(child_nominal_values) == 2

        numerator = self._get_nominal_value_for_sum_subexpression(
            child_nominal_values[0]
        )
        denominator = self._get_nominal_value_for_sum_subexpression(
            child_nominal_values[1]
        )
        if denominator == 0:
            # Assign a nominal value of 1 so that we can continue
            denominator = 1
            # Log a warning for the user
            _log.debug(
                "Nominal value of 0 found in denominator of division expression. "
                "Assigning a value of 1. You should check you scaling factors and models to "
                "ensure there are no values of 0 that can appear in these functions."
            )
        return [numerator / denominator]

    def _get_nominal_value_for_power(self, node, child_nominal_values):
        assert len(child_nominal_values) == 2

        # Use the absolute value of the base term to avoid possible complex numbers
        base = abs(
            self._get_nominal_value_for_sum_subexpression(child_nominal_values[0])
        )
        exponent = self._get_nominal_value_for_sum_subexpression(
            child_nominal_values[1]
        )

        return [base**exponent]

    def _get_nominal_value_single_child(self, node, child_nominal_values):
        assert len(child_nominal_values) == 1
        return child_nominal_values[0]

    def _get_nominal_value_abs(self, node, child_nominal_values):
        assert len(child_nominal_values) == 1
        return [abs(i) for i in child_nominal_values[0]]

    def _get_nominal_value_negation(self, node, child_nominal_values):
        assert len(child_nominal_values) == 1
        return [-i for i in child_nominal_values[0]]

    def _get_nominal_value_for_unary_function(self, node, child_nominal_values):
        assert len(child_nominal_values) == 1
        func_name = node.getname()
        # TODO: Some of these need the absolute value of the nominal value (e.g. sqrt)
        func_nominal = self._get_nominal_value_for_sum_subexpression(
            child_nominal_values[0]
        )
        func = getattr(math, func_name)
        try:
            return [func(func_nominal)]
        except ValueError:
            raise ValueError(
                f"Evaluation error occurred when getting nominal value in {func_name} "
                f"expression with input {func_nominal}. You should check you scaling factors "
                f"and model to address any numerical issues or scale this constraint manually."
            )

    def _get_nominal_value_expr_if(self, node, child_nominal_values):
        assert len(child_nominal_values) == 3
        return child_nominal_values[1] + child_nominal_values[2]

    def _get_nominal_value_external_function(self, node, child_nominal_values):
        # First, need to get expected magnitudes of input terms, which may be sub-expressions
        input_mag = []
        for i in child_nominal_values:
            if isinstance(i[0], str):
                # Sometimes external functions might have string arguments
                # Check here, and return the string if true
                input_mag.append(i[0])
            else:
                input_mag.append(self._get_nominal_value_for_sum_subexpression(i))

        # Next, create a copy of the external function with expected magnitudes as inputs
        newfunc = node.create_node_with_local_data(input_mag)

        # Evaluate new function and return the absolute value
        return [pyo.value(newfunc)]

    node_type_method_map = {
        EXPR.EqualityExpression: _get_nominal_value_for_sum,
        EXPR.InequalityExpression: _get_nominal_value_for_sum,
        EXPR.RangedExpression: _get_nominal_value_for_sum,
        EXPR.SumExpression: _get_nominal_value_for_sum,
        EXPR.NPV_SumExpression: _get_nominal_value_for_sum,
        EXPR.ProductExpression: _get_nominal_value_for_product,
        EXPR.MonomialTermExpression: _get_nominal_value_for_product,
        EXPR.NPV_ProductExpression: _get_nominal_value_for_product,
        EXPR.DivisionExpression: _get_nominal_value_for_division,
        EXPR.NPV_DivisionExpression: _get_nominal_value_for_division,
        EXPR.PowExpression: _get_nominal_value_for_power,
        EXPR.NPV_PowExpression: _get_nominal_value_for_power,
        EXPR.NegationExpression: _get_nominal_value_negation,
        EXPR.NPV_NegationExpression: _get_nominal_value_negation,
        EXPR.AbsExpression: _get_nominal_value_abs,
        EXPR.NPV_AbsExpression: _get_nominal_value_abs,
        EXPR.UnaryFunctionExpression: _get_nominal_value_for_unary_function,
        EXPR.NPV_UnaryFunctionExpression: _get_nominal_value_for_unary_function,
        EXPR.Expr_ifExpression: _get_nominal_value_expr_if,
        EXPR.ExternalFunctionExpression: _get_nominal_value_external_function,
        EXPR.NPV_ExternalFunctionExpression: _get_nominal_value_external_function,
        EXPR.LinearExpression: _get_nominal_value_for_sum,
    }

    def exitNode(self, node, data):
        """Callback for :class:`pyomo.core.current.StreamBasedExpressionVisitor`. This
        method is called when moving back up the tree in a depth first search."""

        # first check if the node is a leaf
        nodetype = type(node)

        if nodetype in native_types:
            return [node]

        node_func = self.node_type_method_map.get(nodetype, None)
        if node_func is not None:
            return node_func(self, node, data)

        elif not node.is_expression_type():
            # this is a leaf, but not a native type
            if nodetype is _PyomoUnit:
                return [1]
            else:
                return self._get_magnitude_base_type(node)
                # might want to add other common types here

        # not a leaf - check if it is a named expression
        if (
            hasattr(node, "is_named_expression_type")
            and node.is_named_expression_type()
        ):
            return self._get_nominal_value_single_child(node, data)

        raise TypeError(
            f"An unhandled expression node type: {str(nodetype)} was encountered while "
            f"retrieving the scaling factor of expression {str(node)}"
        )


def set_constraint_scaling_max_magnitude(
    component, warning: bool = True, overwrite: bool = False, descend_into: bool = True
):
    """
    Set scaling factors for constraints using maximum expected magnitude of additive terms in expression.
    Scaling factor for constraints will be 1 / max(abs(nominal value)).

    Args:
        component: a Pyomo component to set constraint scaling factors for.
        warning: bool indicating whether to log a warning if a missing variable scaling factor is
            found (default=True).
        overwrite: bool indicating whether to overwrite existing scaling factors (default=False).
        descend_into: bool indicating whether function should descend into child Blocks
            if component is a Pyomo Block (default=True).

    Returns:
        None
    """

    def _set_sf_max_mag(c):
        nominal = NominalValueExtractionVisitor(warning=warning).walk_expression(c.expr)
        # 0 terms will never be the largest absolute magnitude, so we can ignore them
        max_mag = max(abs(i) for i in nominal)
        set_scaling_factor(c, max_mag, overwrite=overwrite)

    if isinstance(component, pyo.Block):
        # Iterate over all constraint datas and call this method on each
        for c in component.component_data_objects(
            pyo.Constraint, descend_into=descend_into
        ):
            set_constraint_scaling_max_magnitude(
                c, warning=warning, overwrite=overwrite
            )
    elif component.is_indexed():
        for i in component:
            _set_sf_max_mag(component[i])
    else:
        _set_sf_max_mag(component)


def set_constraint_scaling_min_magnitude(
    component, warning: bool = True, overwrite: bool = False, descend_into: bool = True
):
    """
    Set scaling factors for constraints using minimum expected magnitude of additive terms in expression.
    Scaling factor for constraints will be 1 / min(abs(nominal value)).

    Args:
        component: a Pyomo component to set constraint scaling factors for.
        warning: bool indicating whether to log a warning if a missing variable scaling factor is
            found (default=True).
        overwrite: bool indicating whether to overwrite existing scaling factors (default=False).
        descend_into: bool indicating whether function should descend into child Blocks
            if component is a Pyomo Block (default=True).

    Returns:
        None
    """

    def _set_sf_min_mag(c):
        nominal = NominalValueExtractionVisitor(warning=warning).walk_expression(c.expr)
        # Ignore any 0 terms - we will assume they do not contribute to scaling
        min_mag = min(abs(i) for i in [j for j in nominal if j != 0])
        set_scaling_factor(c, min_mag, overwrite=overwrite)

    if isinstance(component, pyo.Block):
        # Iterate over all constraint datas and call this method on each
        for c in component.component_data_objects(
            pyo.Constraint, descend_into=descend_into
        ):
            set_constraint_scaling_min_magnitude(
                c, warning=warning, overwrite=overwrite
            )
    elif component.is_indexed():
        for i in component:
            _set_sf_min_mag(component[i])
    else:
        _set_sf_min_mag(component)


def set_constraint_scaling_harmonic_magnitude(
    component, warning: bool = True, overwrite: bool = False, descend_into: bool = True
):
    """
    Set scaling factors for constraints using the harmonic sum of the expected magnitude of
    additive terms in expression. Scaling factor for constraints will be 1 / sum(1/abs(nominal value)).

    Args:
        component: a Pyomo component to set constraint scaling factors for.
        warning: bool indicating whether to log a warning if a missing variable scaling factor is
            found (default=True).
        overwrite: bool indicating whether to overwrite existing scaling factors (default=False).
        descend_into: bool indicating whether function should descend into child Blocks
            if component is a Pyomo Block (default=True).

    Returns:
        None
    """

    def _set_sf_har_mag(c):
        nominal = NominalValueExtractionVisitor(warning=warning).walk_expression(c.expr)
        # Ignore any 0 terms - we will assume they do not contribute to scaling
        harm_sum = sum(1 / abs(i) for i in [j for j in nominal if j != 0])
        set_scaling_factor(c, harm_sum, overwrite=overwrite)

    if isinstance(component, pyo.Block):
        # Iterate over all constraint datas and call this method on each
        for c in component.component_data_objects(
            pyo.Constraint, descend_into=descend_into
        ):
            set_constraint_scaling_harmonic_magnitude(
                c, warning=warning, overwrite=overwrite
            )
    elif component.is_indexed():
        for i in component:
            _set_sf_har_mag(component[i])
    else:
        _set_sf_har_mag(component)


def report_scaling_issues(
    blk,
    ostream: "Stream" = None,
    prefix: str = "",
    large: float = 1e4,
    small: float = 1e-3,
    zero: float = 1e-10,
    descend_into: bool = True,
    include_fixed: bool = False,
):
    """
    Write a report on potential scaling issues to a stream.

    Args:
        blk: block to check for scaling issues
        ostream: stream object to write results to (default=stdout)
        prefix: string to prefix output with
        large: Magnitude that is considered to be too large
        small: Magnitude that is considered to be too small
        zero: Magnitude that is considered to be zero, variables with a value of
            zero are okay, and not reported.
        descend_into: bool indicating whether to check constraints in sub-blocks
        include_fixed: bool indicating whether to include fixed Vars in list

    Returns:
        None
    """
    if ostream is None:
        ostream = sys.stdout

    # Write output
    max_str_length = 84
    tab = " " * 4
    ostream.write("\n" + "=" * max_str_length + "\n")
    ostream.write(f"{prefix}Potential Scaling Issues")
    ostream.write("\n" * 2)

    ostream.write(f"{prefix}{tab}Unscaled Variables")
    ostream.write("\n" * 2)
    for v in unscaled_variables_generator(blk, descend_into, include_fixed):
        ostream.write(f"{prefix}{tab*2}{v.name}")
    ostream.write("\n" * 2)
    ostream.write(f"{prefix}{tab}Badly Scaled Variables")
    ostream.write("\n" * 2)
    for v in badly_scaled_var_generator(
        blk, large, small, zero, descend_into, include_fixed
    ):
        ostream.write(f"{prefix}{tab*2}{v[0].name}: {v[1]}")
    ostream.write("\n" * 2)
    ostream.write(f"{prefix}{tab}Unscaled Constraints")
    ostream.write("\n" * 2)
    for c in unscaled_constraints_generator(blk, descend_into):
        ostream.write(f"{prefix}{tab * 2}{c.name}")
    ostream.write("\n")
    ostream.write("\n" + "=" * max_str_length + "\n")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

import logging.config
import json
import os

import pyomo.common.config
from pyomo.common.config import Bool

_log = logging.getLogger(__name__)
# Default release version if no options provided for get-extensions
default_binary_release = "3.4.2"
# Where to download releases from get-extensions
release_base_url = "https://github.com/IDAES/idaes-ext/releases/download"
# Where to get release checksums
release_checksum_url = (
    "https://raw.githubusercontent.com/IDAES/idaes-ext/main/releases/sha256sum_{}.txt"
)
# This is a list of platforms with builds
base_platforms = (
    "darwin-aarch64",
    "darwin-x86_64",
    "el7-x86_64",
    "el8-x86_64",
    "el8-aarch64",
    "ubuntu1804-x86_64",
    "ubuntu1804-aarch64",
    "ubuntu2004-x86_64",
    "ubuntu2004-aarch64",
    "ubuntu2204-x86_64",
    "ubuntu2204-aarch64",
    "windows-x86_64",
)
# Map some platform names to others for get-extensions
binary_distro_map = {
    "macos": "darwin",
    "el9": "ubuntu2204",
    "rhel7": "el7",
    "rhel8": "el8",
    "rhel9": "ubuntu2204",
    "scientific7": "el7",
    "centos7": "el7",
    "centos8": "el8",
    "rocky8": "el8",
    "rocky9": "ubuntu2204",
    "almalinux8": "el8",
    "almalinux9": "ubuntu2204",
    "debian9": "el7",
    "debian10": "el8",
    "debian11": "ubuntu2004",
    "linuxmint20": "ubuntu2004",
    "kubuntu1804": "ubuntu1804",
    "kubuntu2004": "ubuntu2004",
    "kubuntu2204": "ubuntu2204",
    "xubuntu1804": "ubuntu1804",
    "xubuntu2004": "ubuntu2004",
    "xubuntu2204": "ubuntu2204",
    "pop22": "ubuntu2204",
}
# Machine map
binary_arch_map = {
    "x64": "x86_64",
    "intel64": "x86_64",
    "amd64": "x86_64",
    "arm64": "aarch64",
}
# Set of extra binary packages and basic build platform where available
extra_binaries = {
    "petsc": base_platforms,
}
# Store the original environment variable values so we can revert changes
orig_environ = {
    "PATH": os.environ.get("PATH", ""),
    "LD_LIBRARY_PATH": os.environ.get("LD_LIBRARY_PATH", ""),
    "DYLD_LIBRARY_PATH": os.environ.get("DYLD_LIBRARY_PATH", ""),
}
# Define set of default units of measurement for common derived quantities
default_uom = {
    "pressure": "Pa",
    "energy": "J",
    "energy_mol": "J/mol",
    "energy_mass": "J/kg",
    "entropy": "J/K",
    "entropy_mol": "J/mol/K",
    "entropy_mass": "J/kg/K",
    "power": "W",
}


def canonical_arch(arch):
    """Get the official machine type in {x86_64, aarch64} if possible, otherwise
    just return arch.lower().

    Args:
        arch (str): machine type string usually from platform.machine()

    Returns (str):
        Canonical machine type used by the binary package names.
    """
    arch = arch.lower()
    return binary_arch_map.get(arch, arch)


def canonical_distro(dist):
    """Get the official distro name if possible, otherwise just return
    dist.lower(). Distro is used loosely here and includes Windows, Darwin
    (macOS), and other OSs in addition to Linux.

    Args:
        arch (str): machine type string usually from platform.machine()

    Returns (str):
        Canonical machine type used by the binary package names.
    """
    dist = dist.lower()
    return binary_distro_map.get(dist, dist)


class ConfigBlockJSONEncoder(json.JSONEncoder):
    """This class handles non-serializable objects that may appear in the IDAES
    ConfigBlock. For now this is only set objects.
    """

    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return obj


def _new_idaes_config_block():
    """The idaes configuration is stored in a Pyomo ConfigBlock created by this.
    This function is called when importing IDAES.  This function should only be
    called in ``__init__.py`` for ``idaes``.  Calling it anywhere else will cause
    the idaes configuration system to function improperly.
    """
    cfg = pyomo.common.config.ConfigBlock("idaes", implicit=True)
    cfg.declare(
        "warning_to_exception",
        pyomo.common.config.ConfigValue(
            domain=bool,
            default=False,
            description="Convert any logged warnings or errors to exceptions",
            doc="Convert any logged warnings or errors to exceptions",
        ),
    )
    cfg.declare(
        "deprecation_to_exception",
        pyomo.common.config.ConfigValue(
            domain=bool,
            default=False,
            description="Convert any logged deprecation warnings to exceptions",
            doc="Convert any logged deprecation warnings to exceptions",
        ),
    )
    cfg.declare(
        "logging",
        pyomo.common.config.ConfigBlock(
            implicit=True,
            description="Logging configuration dictionary",
            doc="This stores the logging configuration. See the Python "
            "logging.config.dictConfig() documentation for details.",
        ),
    )
    cfg["logging"].declare(
        "version", pyomo.common.config.ConfigValue(domain=int, default=1)
    )
    cfg["logging"].declare(
        "disable_existing_loggers",
        pyomo.common.config.ConfigValue(domain=bool, default=False),
    )
    cfg["logging"].declare("formatters", pyomo.common.config.ConfigBlock(implicit=True))
    cfg["logging"]["formatters"].declare(
        "default_format", pyomo.common.config.ConfigBlock(implicit=True)
    )
    cfg["logging"]["formatters"]["default_format"].declare(
        "format",
        pyomo.common.config.ConfigValue(
            domain=str, default="%(asctime)s [%(levelname)s] %(name)s: %(message)s"
        ),
    )
    cfg["logging"]["formatters"]["default_format"].declare(
        "datefmt",
        pyomo.common.config.ConfigValue(domain=str, default="%Y-%m-%d %H:%M:%S"),
    )
    cfg["logging"]["formatters"].declare(
        "blank_format", pyomo.common.config.ConfigBlock(implicit=True)
    )
    cfg["logging"]["formatters"]["blank_format"].declare(
        "format",
        pyomo.common.config.ConfigValue(domain=str, default="%(message)s"),
    )
    cfg["logging"].declare("handlers", pyomo.common.config.ConfigBlock(implicit=True))
    cfg["logging"]["handlers"].declare(
        "console", pyomo.common.config.ConfigBlock(implicit=True)
    )
    cfg["logging"]["handlers"]["console"].declare(
        "class",
        pyomo.common.config.ConfigValue(domain=str, default="logging.StreamHandler"),
    )
    cfg["logging"]["handlers"]["console"].declare(
        "formatter",
        pyomo.common.config.ConfigValue(domain=str, default="default_format"),
    )
    cfg["logging"]["handlers"]["console"].declare(
        "stream",
        pyomo.common.config.ConfigValue(domain=str, default="ext://sys.stdout"),
    )
    cfg["logging"]["handlers"].declare(
        "console_blank", pyomo.common.config.ConfigBlock(implicit=True)
    )
    cfg["logging"]["handlers"]["console_blank"].declare(
        "class",
        pyomo.common.config.ConfigValue(domain=str, default="logging.StreamHandler"),
    )
    cfg["logging"]["handlers"]["console_blank"].declare(
        "formatter",
        pyomo.common.config.ConfigValue(domain=str, default="blank_format"),
    )
    cfg["logging"]["handlers"]["console_blank"].declare(
        "stream",
        pyomo.common.config.ConfigValue(domain=str, default="ext://sys.stdout"),
    )
    cfg["logging"].declare(
        "loggers",
        pyomo.common.config.ConfigValue(
            domain=dict,
            default={  # the period in the logger names is trouble for ConfigBlock
                "idaes": {"level": "INFO", "propagate": True, "handlers": ["console"]},
                "idaes.solve": {"propagate": False, "handlers": ["console"]},
                "idaes.init": {"propagate": False, "handlers": ["console"]},
                "idaes.model": {"propagate": False, "handlers": ["console"]},
                "idaes.helmholtz_parameters": {
                    "propagate": False,
                    "handlers": ["console_blank"],
                },
            },
        ),
    )
    cfg.declare(
        "properties",
        pyomo.common.config.ConfigBlock(
            implicit=False,
            description="Physical-property-related options",
            doc="Physical-property-related options",
        ),
    )
    cfg["properties"].declare(
        "helmholtz",
        pyomo.common.config.ConfigBlock(
            implicit=False,
            description="Helmholtz equation of state configuration block",
            doc="Helmholtz equation of state configuration block",
        ),
    )
    cfg["properties"]["helmholtz"].declare(
        "parameter_file_path",
        pyomo.common.config.ConfigValue(
            domain=str,
            default=None,
            description="Helmholtz parameter file path",
            doc="Helmholtz parameter file path, if None use default based on "
            "IDAES install location",
        ),
    )

    cfg.declare(
        "ipopt",
        pyomo.common.config.ConfigBlock(
            implicit=False,
            description="Default config for 'ipopt' solver",
            doc="Default config for 'ipopt' solver",
        ),
    )
    cfg["ipopt"].declare(
        "options",
        pyomo.common.config.ConfigBlock(
            implicit=True,
            description="Default solver options for 'ipopt'",
            doc="Default solver options for 'ipopt' solver",
        ),
    )

    cfg["ipopt"]["options"].declare(
        "nlp_scaling_method",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="gradient-based",
            description="Ipopt NLP scaling method",
            doc="Ipopt NLP scaling method",
        ),
    )

    cfg["ipopt"]["options"].declare(
        "tol",
        pyomo.common.config.ConfigValue(
            domain=float,
            default=1e-6,
            description="Ipopt tol option",
            doc="Ipopt tol option",
        ),
    )

    cfg["ipopt"]["options"].declare(
        "max_iter",
        pyomo.common.config.ConfigValue(
            domain=int,
            default=200,
            description="Ipopt max_iter option",
            doc="Ipopt max_iter option",
        ),
    )

    cfg.declare(
        "ipopt_v2",
        pyomo.common.config.ConfigBlock(
            implicit=False,
            description="Default config for 'ipopt' solver",
        ),
    )
    cfg["ipopt_v2"].declare(
        "options",
        pyomo.common.config.ConfigBlock(
            implicit=True,
            description="Default solver options for 'ipopt'",
        ),
    )

    cfg["ipopt_v2"]["options"].declare(
        "nlp_scaling_method",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="gradient-based",
            description="Ipopt NLP scaling method",
        ),
    )

    cfg["ipopt_v2"]["options"].declare(
        "tol",
        pyomo.common.config.ConfigValue(
            domain=float,
            default=1e-6,
            description="Ipopt tol option",
        ),
    )

    cfg["ipopt_v2"]["options"].declare(
        "max_iter",
        pyomo.common.config.ConfigValue(
            domain=int,
            default=200,
            description="Ipopt max_iter option",
        ),
    )

    cfg["ipopt_v2"]["options"].declare(
        "linear_solver",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="ma57",
            description="Linear solver to be used by IPOPT",
        ),
    )

    cfg["ipopt_v2"].declare(
        "writer_config",
        pyomo.common.config.ConfigBlock(
            implicit=True,
            description="Default writer configuration for 'ipopt'",
        ),
    )

    # TODO: Remember to update BTInitializer to use get_solver once scaling tools are deployed.
    cfg["ipopt_v2"]["writer_config"].declare(
        "scale_model",
        pyomo.common.config.ConfigValue(
            domain=Bool,
            default=False,  # TODO: Change to true once transition complete
            description="Whether to apply model scaling in writer",
        ),
    )

    cfg["ipopt_v2"]["writer_config"].declare(
        "linear_presolve",
        pyomo.common.config.ConfigValue(
            domain=Bool,
            default=True,
            description="Whether to apply linear presolve in writer",
        ),
    )

    cfg.declare(
        "ipopt_l1",
        pyomo.common.config.ConfigBlock(
            implicit=False,
            description="Default config for 'ipopt_l1' solver",
            doc="Default config for 'ipopt_l1' solver",
        ),
    )

    cfg["ipopt_l1"].declare(
        "options",
        pyomo.common.config.ConfigBlock(
            implicit=True,
            description="Default solver options for 'ipopt_l1'",
            doc="Default solver options for 'ipopt_l1' solver",
        ),
    )

    cfg["ipopt_l1"]["options"].declare(
        "nlp_scaling_method",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="gradient-based",
            description="Ipopt_l1 NLP scaling method",
            doc="Ipopt_l1 NLP scaling method",
        ),
    )

    cfg["ipopt_l1"]["options"].declare(
        "tol",
        pyomo.common.config.ConfigValue(
            domain=float,
            default=1e-6,
            description="Ipopt_l1 tol option",
            doc="Ipopt_l1 tol option",
        ),
    )

    cfg["ipopt_l1"]["options"].declare(
        "max_iter",
        pyomo.common.config.ConfigValue(
            domain=int,
            default=200,
            description="Ipopt_l1 max_iter option",
            doc="Ipopt_l1 max_iter option",
        ),
    )

    cfg.declare(
        "petsc_ts",
        pyomo.common.config.ConfigBlock(
            implicit=False,
            description="Default config for 'petsc_ts' solver",
            doc="Default config for 'petsc_ts' solver",
        ),
    )

    cfg["petsc_ts"].declare(
        "options",
        pyomo.common.config.ConfigBlock(
            implicit=True,
            description="Default solver options for 'petsc_ts'",
            doc="Default solver options for 'petsc_ts' solver",
        ),
    )

    cfg["petsc_ts"]["options"].declare(
        "--ts_save_trajectory",
        pyomo.common.config.ConfigValue(
            domain=int,
            default=1,
            description="Save the trajectory data from PETSc",
            doc="Save the trajectory data from PETSc",
        ),
    )

    cfg["petsc_ts"]["options"].declare(
        "--ts_max_snes_failures",
        pyomo.common.config.ConfigValue(
            domain=int,
            default=200,
            description="Number of nonlinear solver failures before giving up",
            doc="Number of nonlinear solver failures before giving up",
        ),
    )

    cfg["petsc_ts"]["options"].declare(
        "--ts_max_reject",
        pyomo.common.config.ConfigValue(
            domain=int,
            default=20,
            description="Maximum number of time steps to reject",
            doc="Maximum number of time steps to reject",
        ),
    )

    cfg["petsc_ts"]["options"].declare(
        "--ts_type",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="beuler",
            description="TS solver to use",
            doc="TS solver to use",
        ),
    )

    cfg["petsc_ts"]["options"].declare(
        "--ts_adapt_type",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="basic",
            description="TS adaptive step size method to use",
            doc="TS adaptive step size method to use",
        ),
    )

    cfg["petsc_ts"]["options"].declare(
        "--ts_exact_final_time",
        pyomo.common.config.ConfigValue(
            domain=str,
            default="matchstep",
            description="How to handle the final time step",
            doc="How to handle the final time step",
        ),
    )

    cfg.declare(
        "default_solver",
        pyomo.common.config.ConfigValue(
            default="ipopt",
            domain=str,
            description="Default solver.  See Pyomo's SolverFactory for details.",
            doc="Default solver.  See Pyomo's SolverFactory for details.",
        ),
    )

    cfg.declare(
        "use_idaes_solvers",
        pyomo.common.config.ConfigValue(
            default=True,
            domain=bool,
            description="If True, search the IDAES bin directory for executables"
            " first; otherwise, use IDAES bin directory as last resort.",
            doc="If True the IDAES bin directory will be searched for executables "
            "first, which will result in the solvers installed by IDAES being "
            "used in preference to solvers installed on the machine.  If False, "
            "IDAES will only fall back on solvers installed in the IDAES bin "
            "directory if they are not otherwise available.",
        ),
    )

    cfg.declare(
        "use_idaes_solver_config",
        pyomo.common.config.ConfigValue(
            default=False,
            domain=bool,
            description="If True, use the configure IDAES solver default options.",
            doc="If True, use the configure IDAES solver default options.",
        ),
    )

    cfg.declare(
        "valid_logger_tags",
        pyomo.common.config.ConfigValue(
            default=set(
                [
                    "framework",
                    "model",
                    "flowsheet",
                    "unit",
                    "control_volume",
                    "properties",
                    "reactions",
                    "ui",
                ]
            ),
            domain=set,
            description="List of valid logger tags",
        ),
    )

    cfg.declare(
        "logger_tags",
        pyomo.common.config.ConfigValue(
            default=set(
                [
                    "framework",
                    "model",
                    "flowsheet",
                    "unit",
                    "control_volume",
                    "properties",
                    "reactions",
                ]
            ),
            domain=set,
            description="List of logger tags to allow",
        ),
    )

    cfg.declare(
        "logger_capture_solver",
        pyomo.common.config.ConfigValue(
            default=True,
            description="Solver output captured by logger?",
        ),
    )

    cfg.declare(
        "reporting_units",
        pyomo.common.config.ConfigValue(
            default=default_uom,
            description="Units of measurement for reporting quantities",
        ),
    )
    return cfg


def read_config(val, cfg):
    """Read either a JSON formatted config file or a configuration dictionary.
    Args:
        val: dict, ConfigDict, or file to read
        cfg: config block
    Returns:
        None
    """
    config_file = None
    if val is None:
        return
    elif isinstance(val, (dict, pyomo.common.config.ConfigBlock)):
        pass  # don't worry this catches ConfigBlock too it seems
    else:
        config_file = val
        try:
            with open(config_file, "r") as f:
                val = json.load(f)
        except IOError:  # don't require config file
            _log.debug(f"Config file {config_file} not found (this is okay)")
            return
    cfg.set_value(val)
    if config_file is not None:
        _log.debug(f"Read config {config_file}")
    reconfig(cfg)


def write_config(path, cfg=None):
    if cfg is None:
        cfg = _new_idaes_config_block()
    with open(path, "w") as f:
        json.dump(cfg.value(), f, cls=ConfigBlockJSONEncoder, indent=4)


class _WarningToExceptionFilter(logging.Filter):
    """Filter applied to IDAES loggers returned by this module."""

    @staticmethod
    def filter(record):
        if record.levelno >= logging.WARNING:
            raise RuntimeError(f"Logged Warning converted to exception: {record.msg}")


class _DeprecationToExceptionFilter(logging.Filter):
    """Filter applied to IDAES loggers returned by this module."""

    @staticmethod
    def filter(record):
        if record.levelno >= logging.WARNING:
            if "deprecat" in record.msg.lower():
                raise RuntimeError(
                    f"Logged deprecation converted to exception: {record.msg}"
                )


def reconfig(cfg):
    logging.config.dictConfig(cfg.logging.value())
    _log = logging.getLogger("idaes")
    if cfg.deprecation_to_exception:
        _log.addFilter(_DeprecationToExceptionFilter)
    else:
        _log.removeFilter(_DeprecationToExceptionFilter)
    if cfg.warning_to_exception:
        _log.addFilter(_WarningToExceptionFilter)
    else:
        _log.removeFilter(_WarningToExceptionFilter)
    setup_environment(bin_directory, cfg.use_idaes_solvers)


def create_dir(d):
    """Create a directory if it doesn't exist.

    Args:
        d(str): directory path to create

    Returns:
        None
    """
    if os.path.exists(d):
        return
    else:
        os.mkdir(d)


def get_data_directory():
    """Return the standard data directory for idaes, based on the OS."""
    if "IDAES_DATA" in os.environ:
        data_directory = os.environ["IDAES_DATA"]
    else:
        try:
            if os.name == "nt":  # Windows
                data_directory = os.path.join(os.environ["LOCALAPPDATA"], "idaes")
            else:  # any other OS
                data_directory = os.path.join(os.environ["HOME"], ".idaes")
        except AttributeError:
            data_directory = None
    if data_directory is None or not os.path.isdir(os.path.dirname(data_directory)):
        _log.warning(
            "IDAES data directory location does not exist: "
            f"{os.path.dirname(data_directory)}"
        )
        data_directory = None
    # Standard location for executable binaries.
    if data_directory is not None:
        bin_directory = os.path.join(data_directory, "bin")
    else:
        bin_directory = None
    # Standard location for testing files
    if data_directory is not None:
        testing_directory = os.path.join(data_directory, "testing")
    else:
        testing_directory = None

    return data_directory, bin_directory, testing_directory


def setup_environment(bin_directory, use_idaes_solvers):
    """
    Set environment variables for the IDAES session.

    Args:
        bin_directory: directory to find idaes libraries and executables
        use_idaes_solvers: If true look first in the idaes bin directory for
                           executables if false look last in the idaes bin
                           directory for executables.

    Returns:
        None
    """
    if bin_directory is None:
        return
    oe = orig_environ
    if use_idaes_solvers:
        os.environ["PATH"] = os.pathsep.join([bin_directory, oe.get("PATH", "")])
    else:
        os.environ["PATH"] = os.pathsep.join([oe.get("PATH", ""), bin_directory])
    if os.name != "nt":  # If not Windows set lib search path, Windows uses PATH
        os.environ["LD_LIBRARY_PATH"] = os.pathsep.join(
            [oe.get("LD_LIBRARY_PATH", ""), bin_directory]
        )
        # This is for macOS, but won't hurt other UNIX
        os.environ["DYLD_LIBRARY_PATH"] = os.pathsep.join(
            [oe.get("DYLD_LIBRARY_PATH", ""), bin_directory]
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Plate Heat Exchanger unit model.
Author: Akula Paul, Andrew Lee
"""

import pytest

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    Expression,
    Param,
    units as pyunits,
    value,
    Var,
)
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent

from idaes.core import FlowsheetBlock
from idaes.models.unit_models.heat_exchanger_ntu import (
    HeatExchangerNTU as HXNTU,
    HXNTUInitializer,
)

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models_extra.column_models.properties.MEA_solvent import (
    configuration as aqueous_mea,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver
from idaes.core.util.testing import initialization_tester, PhysicalParameterTestBlock
from idaes.core.util.exceptions import ConfigurationError, InitializationError
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
    InitializationStatus,
)
from idaes.core.util import DiagnosticsToolbox


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_bad_option():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    with pytest.raises(KeyError):
        m.fs.unit = HXNTU(**{"I'm a bad option": "hot"})


@pytest.mark.unit
def test_bad_option2():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    with pytest.raises(
        ConfigurationError, match="cold_side_name cannot be 'hot_side'."
    ):
        m.fs.unit = HXNTU(cold_side_name="hot_side")


@pytest.mark.unit
def test_bad_option3():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    with pytest.raises(
        ConfigurationError, match="hot_side_name cannot be 'cold_side'."
    ):
        m.fs.unit = HXNTU(hot_side_name="cold_side")


@pytest.mark.unit
def test_bad_option4():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    with pytest.raises(
        ConfigurationError, match="cold_side_name cannot be 'cold_side'."
    ):
        m.fs.unit = HXNTU(cold_side_name="cold_side")


@pytest.mark.unit
def test_bad_option5():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    with pytest.raises(ConfigurationError, match="hot_side_name cannot be 'hot_side'."):
        m.fs.unit = HXNTU(hot_side_name="hot_side")


@pytest.mark.unit
def test_hot_and_cold_names_same():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    with pytest.raises(
        NameError,
        match="HeatExchanger hot and cold side cannot have the same name 'shell'.",
    ):
        m.fs.unit = HXNTU(hot_side_name="shell", cold_side_name="shell")


@pytest.mark.unit
def test_hot_side_name_clash():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = PhysicalParameterTestBlock()

    with pytest.raises(
        ValueError,
        match="fs.unit could not assign hot side alias "
        "build as an attribute of that name already "
        "exists.",
    ):
        m.fs.unit = HXNTU(
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties},
            hot_side_name="build",
        )


@pytest.mark.unit
def test_cold_side_name_clash():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = PhysicalParameterTestBlock()

    with pytest.raises(
        ValueError,
        match="fs.unit could not assign cold side alias "
        "build as an attribute of that name already "
        "exists.",
    ):
        m.fs.unit = HXNTU(
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties},
            cold_side_name="build",
        )


@pytest.mark.unit
def test_user_names():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = HXNTU(
        hot_side_name="shell",
        cold_side_name="tube",
        shell={"property_package": m.fs.properties},
        tube={"property_package": m.fs.properties},
    )

    assert m.fs.unit.config.hot_side.property_package is m.fs.properties
    assert m.fs.unit.config.cold_side.property_package is m.fs.properties

    assert m.fs.unit.shell is m.fs.unit.hot_side
    assert m.fs.unit.tube is m.fs.unit.cold_side

    assert m.fs.unit.shell_inlet is m.fs.unit.hot_side_inlet
    assert m.fs.unit.tube_inlet is m.fs.unit.cold_side_inlet
    assert m.fs.unit.shell_outlet is m.fs.unit.hot_side_outlet
    assert m.fs.unit.tube_outlet is m.fs.unit.cold_side_outlet


# -----------------------------------------------------------------------------
class TestHXNTU(object):
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.hotside_properties = GenericParameterBlock(**aqueous_mea)
        m.fs.coldside_properties = GenericParameterBlock(**aqueous_mea)

        m.fs.unit = HXNTU(
            hot_side={
                "property_package": m.fs.hotside_properties,
                "has_pressure_change": True,
            },
            cold_side={
                "property_package": m.fs.coldside_properties,
                "has_pressure_change": True,
            },
        )

        # Hot fluid
        m.fs.unit.hot_side_inlet.flow_mol[0].fix(60.54879)
        m.fs.unit.hot_side_inlet.temperature[0].fix(392.23)
        m.fs.unit.hot_side_inlet.pressure[0].fix(202650)
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "CO2"].fix(0.0158)
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "H2O"].fix(0.8747)
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "MEA"].fix(0.1095)

        # Cold fluid
        m.fs.unit.cold_side_inlet.flow_mol[0].fix(63.01910)
        m.fs.unit.cold_side_inlet.temperature[0].fix(326.36)
        m.fs.unit.cold_side_inlet.pressure[0].fix(202650)
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "CO2"].fix(0.0414)
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "H2O"].fix(0.8509)
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "MEA"].fix(0.1077)

        # Unit design variables
        m.fs.unit.area.fix(100)
        m.fs.unit.heat_transfer_coefficient.fix(200)
        m.fs.unit.effectiveness.fix(0.7)

        m.fs.unit.hot_side.deltaP.fix(-2000)
        m.fs.unit.cold_side.deltaP.fix(-2000)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, model):

        assert hasattr(model.fs.unit, "hot_side_inlet")
        assert len(model.fs.unit.hot_side_inlet.vars) == 4
        assert hasattr(model.fs.unit.hot_side_inlet, "flow_mol")
        assert hasattr(model.fs.unit.hot_side_inlet, "mole_frac_comp")
        assert hasattr(model.fs.unit.hot_side_inlet, "temperature")
        assert hasattr(model.fs.unit.hot_side_inlet, "pressure")

        assert hasattr(model.fs.unit, "hot_side_outlet")
        assert len(model.fs.unit.hot_side_outlet.vars) == 4
        assert hasattr(model.fs.unit.hot_side_outlet, "flow_mol")
        assert hasattr(model.fs.unit.hot_side_outlet, "mole_frac_comp")
        assert hasattr(model.fs.unit.hot_side_outlet, "temperature")
        assert hasattr(model.fs.unit.hot_side_outlet, "pressure")

        assert hasattr(model.fs.unit, "cold_side_inlet")
        assert len(model.fs.unit.cold_side_inlet.vars) == 4
        assert hasattr(model.fs.unit.cold_side_inlet, "flow_mol")
        assert hasattr(model.fs.unit.cold_side_inlet, "mole_frac_comp")
        assert hasattr(model.fs.unit.cold_side_inlet, "temperature")
        assert hasattr(model.fs.unit.cold_side_inlet, "pressure")

        assert hasattr(model.fs.unit, "cold_side_outlet")
        assert len(model.fs.unit.cold_side_outlet.vars) == 4
        assert hasattr(model.fs.unit.cold_side_outlet, "flow_mol")
        assert hasattr(model.fs.unit.cold_side_outlet, "mole_frac_comp")
        assert hasattr(model.fs.unit.cold_side_outlet, "temperature")
        assert hasattr(model.fs.unit.cold_side_outlet, "pressure")

        assert hasattr(model.fs.unit, "heat_duty")
        assert hasattr(model.fs.unit.cold_side, "heat")
        assert hasattr(model.fs.unit.hot_side, "heat")

        assert hasattr(model.fs.unit.cold_side, "deltaP")
        assert hasattr(model.fs.unit.hot_side, "deltaP")

        assert isinstance(model.fs.unit.area, Var)
        assert isinstance(model.fs.unit.heat_transfer_coefficient, Var)
        assert isinstance(model.fs.unit.effectiveness, Var)

        assert isinstance(model.fs.unit.eps_cmin, Param)
        assert value(model.fs.unit.eps_cmin) == 1e-3

        assert isinstance(model.fs.unit.Cmin, Expression)
        assert isinstance(model.fs.unit.Cmax, Expression)
        assert isinstance(model.fs.unit.Cratio, Expression)
        assert isinstance(model.fs.unit.NTU, Expression)

        assert isinstance(model.fs.unit.energy_balance_constraint, Constraint)
        assert isinstance(model.fs.unit.heat_duty_constraint, Constraint)

        assert model.fs.unit.default_initializer is HXNTUInitializer

    @pytest.mark.component
    def test_structural_issues(self, model):
        dt = DiagnosticsToolbox(model)
        dt.assert_no_structural_warnings(
            ignore_evaluation_errors=True,
        )

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, model):
        perf_dict = model.fs.unit._get_performance_contents()

        assert perf_dict is None

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_stream_table_contents(self, model):
        stable = model.fs.unit._get_stream_table_contents()

        expected = {
            "Units": {
                "Total Molar Flowrate": getattr(pyunits.pint_registry, "mole/second"),
                "Total Mole Fraction H2O": getattr(
                    pyunits.pint_registry, "dimensionless"
                ),
                "Total Mole Fraction MEA": getattr(
                    pyunits.pint_registry, "dimensionless"
                ),
                "Total Mole Fraction CO2": getattr(
                    pyunits.pint_registry, "dimensionless"
                ),
                "Temperature": getattr(pyunits.pint_registry, "kelvin"),
                "Pressure": getattr(pyunits.pint_registry, "Pa"),
            },
            "Hot Inlet": {
                "Total Molar Flowrate": pytest.approx(60.549, rel=1e-4),
                "Total Mole Fraction H2O": pytest.approx(0.87470, rel=1e-4),
                "Total Mole Fraction MEA": pytest.approx(0.10950, rel=1e-4),
                "Total Mole Fraction CO2": pytest.approx(0.015800, rel=1e-4),
                "Temperature": pytest.approx(392.23, rel=1e-4),
                "Pressure": pytest.approx(2.0265e05, rel=1e-4),
            },
            "Hot Outlet": {
                "Total Molar Flowrate": pytest.approx(1, rel=1e-4),
                "Total Mole Fraction H2O": pytest.approx(1 / 3, rel=1e-4),
                "Total Mole Fraction MEA": pytest.approx(1 / 3, rel=1e-4),
                "Total Mole Fraction CO2": pytest.approx(1 / 3, rel=1e-4),
                "Temperature": pytest.approx(298.15, rel=1e-4),
                "Pressure": pytest.approx(101325, rel=1e-4),
            },
            "Cold Inlet": {
                "Total Molar Flowrate": pytest.approx(63.019, rel=1e-4),
                "Total Mole Fraction H2O": pytest.approx(0.85090, rel=1e-4),
                "Total Mole Fraction MEA": pytest.approx(0.10770, rel=1e-4),
                "Total Mole Fraction CO2": pytest.approx(0.041400, rel=1e-4),
                "Temperature": pytest.approx(326.36, rel=1e-4),
                "Pressure": pytest.approx(2.0265e05, rel=1e-4),
            },
            "Cold Outlet": {
                "Total Molar Flowrate": pytest.approx(1, rel=1e-4),
                "Total Mole Fraction H2O": pytest.approx(1 / 3, rel=1e-4),
                "Total Mole Fraction MEA": pytest.approx(1 / 3, rel=1e-4),
                "Total Mole Fraction CO2": pytest.approx(1 / 3, rel=1e-4),
                "Temperature": pytest.approx(298.15, rel=1e-4),
                "Pressure": pytest.approx(101325, rel=1e-4),
            },
        }

        assert stable.to_dict() == expected

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(200650, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.pressure[0]
        )
        assert pytest.approx(200650, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.pressure[0]
        )

        assert pytest.approx(343.995, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.temperature[0]
        )
        assert pytest.approx(374.333, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.temperature[0]
        )

        assert pytest.approx(0.015800, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.10950, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.mole_frac_comp[0, "MEA"]
        )
        assert pytest.approx(0.87470, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.mole_frac_comp[0, "H2O"]
        )

        assert pytest.approx(0.041400, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.10770, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.mole_frac_comp[0, "MEA"]
        )
        assert pytest.approx(0.85090, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.mole_frac_comp[0, "H2O"]
        )

        Cmin = value(
            model.fs.unit.hot_side.properties_in[0].flow_mol
            * model.fs.unit.hot_side.properties_in[0].cp_mol
        )
        assert value(model.fs.unit.Cmin[0]) == pytest.approx(Cmin, rel=1e-5)
        assert value(model.fs.unit.Cmax[0]) == pytest.approx(
            value(
                model.fs.unit.cold_side.properties_in[0].flow_mol
                * model.fs.unit.cold_side.properties_in[0].cp_mol
            ),
            rel=1e-5,
        )
        assert value(model.fs.unit.NTU[0]) == pytest.approx(
            value(
                model.fs.unit.area * model.fs.unit.heat_transfer_coefficient[0] / Cmin
            ),
            rel=1e-5,
        )

        assert pytest.approx(0.7 * Cmin * (392.23 - 326.36), rel=1e-5) == value(
            model.fs.unit.heat_duty[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        # Mass conservation test
        assert (
            abs(
                value(
                    model.fs.unit.hot_side_inlet.flow_mol[0]
                    - model.fs.unit.hot_side_outlet.flow_mol[0]
                )
            )
            <= 1e-6
        )

        for j in model.fs.hotside_properties.apparent_species_set:
            assert (
                abs(
                    value(
                        model.fs.unit.hot_side_inlet.flow_mol[0]
                        * model.fs.unit.hot_side_inlet.mole_frac_comp[0, j]
                        - model.fs.unit.hot_side_outlet.flow_mol[0]
                        * model.fs.unit.hot_side_outlet.mole_frac_comp[0, j]
                    )
                )
                <= 1e-6
            )

        assert (
            abs(
                value(
                    model.fs.unit.cold_side_inlet.flow_mol[0]
                    - model.fs.unit.cold_side_outlet.flow_mol[0]
                )
            )
            <= 1e-6
        )

        for j in model.fs.coldside_properties.apparent_species_set:
            assert (
                abs(
                    value(
                        model.fs.unit.cold_side_inlet.flow_mol[0]
                        * model.fs.unit.cold_side_inlet.mole_frac_comp[0, j]
                        - model.fs.unit.cold_side_outlet.flow_mol[0]
                        * model.fs.unit.cold_side_outlet.mole_frac_comp[0, j]
                    )
                )
                <= 1e-6
            )

        # Energy conservation test
        assert (
            abs(
                value(
                    model.fs.unit.hot_side.properties_in[0]._enthalpy_flow_term["Liq"]
                    + model.fs.unit.cold_side.properties_in[0]._enthalpy_flow_term[
                        "Liq"
                    ]
                    - model.fs.unit.hot_side.properties_out[0]._enthalpy_flow_term[
                        "Liq"
                    ]
                    - model.fs.unit.cold_side.properties_out[0]._enthalpy_flow_term[
                        "Liq"
                    ]
                )
            )
            <= 1e-6
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, model):
        dt = DiagnosticsToolbox(model)
        dt.assert_no_numerical_warnings()


class TestInitializers(object):
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.hotside_properties = GenericParameterBlock(**aqueous_mea)
        m.fs.coldside_properties = GenericParameterBlock(**aqueous_mea)

        m.fs.unit = HXNTU(
            hot_side={
                "property_package": m.fs.hotside_properties,
                "has_pressure_change": True,
            },
            cold_side={
                "property_package": m.fs.coldside_properties,
                "has_pressure_change": True,
            },
        )

        # Hot fluid
        m.fs.unit.hot_side_inlet.flow_mol[0].set_value(60.54879)
        m.fs.unit.hot_side_inlet.temperature[0].set_value(392.23)
        m.fs.unit.hot_side_inlet.pressure[0].set_value(202650)
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "CO2"].set_value(0.0158)
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "H2O"].set_value(0.8747)
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "MEA"].set_value(0.1095)

        # Cold fluid
        m.fs.unit.cold_side_inlet.flow_mol[0].set_value(63.01910)
        m.fs.unit.cold_side_inlet.temperature[0].set_value(326.36)
        m.fs.unit.cold_side_inlet.pressure[0].set_value(202650)
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "CO2"].set_value(0.0414)
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "H2O"].set_value(0.8509)
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "MEA"].set_value(0.1077)

        # Unit design variables
        m.fs.unit.area.fix(100)
        m.fs.unit.heat_transfer_coefficient.fix(200)
        m.fs.unit.effectiveness.fix(0.7)

        m.fs.unit.hot_side.deltaP.fix(-2000)
        m.fs.unit.cold_side.deltaP.fix(-2000)

        return m

    @pytest.mark.component
    def test_hx_ntu_initializer(self, model):
        # Need to estimate outlet state values based on heat duty
        initializer = HXNTUInitializer(always_estimate_states=True)
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert pytest.approx(200650, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.pressure[0]
        )
        assert pytest.approx(200650, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.pressure[0]
        )

        assert pytest.approx(343.995, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.temperature[0]
        )
        assert pytest.approx(374.333, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.temperature[0]
        )

        assert pytest.approx(0.015800, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.10950, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.mole_frac_comp[0, "MEA"]
        )
        assert pytest.approx(0.87470, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.mole_frac_comp[0, "H2O"]
        )

        assert pytest.approx(0.041400, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.10770, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.mole_frac_comp[0, "MEA"]
        )
        assert pytest.approx(0.85090, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.mole_frac_comp[0, "H2O"]
        )

        Cmin = value(
            model.fs.unit.hot_side.properties_in[0].flow_mol
            * model.fs.unit.hot_side.properties_in[0].cp_mol
        )
        assert value(model.fs.unit.Cmin[0]) == pytest.approx(Cmin, rel=1e-5)
        assert value(model.fs.unit.Cmax[0]) == pytest.approx(
            value(
                model.fs.unit.cold_side.properties_in[0].flow_mol
                * model.fs.unit.cold_side.properties_in[0].cp_mol
            ),
            rel=1e-5,
        )
        assert value(model.fs.unit.NTU[0]) == pytest.approx(
            value(
                model.fs.unit.area * model.fs.unit.heat_transfer_coefficient[0] / Cmin
            ),
            rel=1e-5,
        )

        assert pytest.approx(0.7 * Cmin * (392.23 - 326.36), rel=1e-5) == value(
            model.fs.unit.heat_duty[0]
        )

        assert not model.fs.unit.hot_side_inlet.flow_mol[0].fixed
        assert not model.fs.unit.hot_side_inlet.temperature[0].fixed
        assert not model.fs.unit.hot_side_inlet.pressure[0].fixed
        assert not model.fs.unit.hot_side_inlet.mole_frac_comp[0, "CO2"].fixed
        assert not model.fs.unit.hot_side_inlet.mole_frac_comp[0, "H2O"].fixed
        assert not model.fs.unit.hot_side_inlet.mole_frac_comp[0, "MEA"].fixed

        assert not model.fs.unit.cold_side_inlet.flow_mol[0].fixed
        assert not model.fs.unit.cold_side_inlet.temperature[0].fixed
        assert not model.fs.unit.cold_side_inlet.pressure[0].fixed
        assert not model.fs.unit.cold_side_inlet.mole_frac_comp[0, "CO2"].fixed
        assert not model.fs.unit.cold_side_inlet.mole_frac_comp[0, "H2O"].fixed
        assert not model.fs.unit.cold_side_inlet.mole_frac_comp[0, "MEA"].fixed

    @pytest.mark.component
    def test_block_triangularization_initializer(self, model):
        initializer = BlockTriangularizationInitializer(constraint_tolerance=2e-5)
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert pytest.approx(200650, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.pressure[0]
        )
        assert pytest.approx(200650, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.pressure[0]
        )

        assert pytest.approx(343.995, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.temperature[0]
        )
        assert pytest.approx(374.333, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.temperature[0]
        )

        assert pytest.approx(0.015800, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.10950, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.mole_frac_comp[0, "MEA"]
        )
        assert pytest.approx(0.87470, rel=1e-5) == value(
            model.fs.unit.hot_side_outlet.mole_frac_comp[0, "H2O"]
        )

        assert pytest.approx(0.041400, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.mole_frac_comp[0, "CO2"]
        )
        assert pytest.approx(0.10770, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.mole_frac_comp[0, "MEA"]
        )
        assert pytest.approx(0.85090, rel=1e-5) == value(
            model.fs.unit.cold_side_outlet.mole_frac_comp[0, "H2O"]
        )

        Cmin = value(
            model.fs.unit.hot_side.properties_in[0].flow_mol
            * model.fs.unit.hot_side.properties_in[0].cp_mol
        )
        assert value(model.fs.unit.Cmin[0]) == pytest.approx(Cmin, rel=1e-5)
        assert value(model.fs.unit.Cmax[0]) == pytest.approx(
            value(
                model.fs.unit.cold_side.properties_in[0].flow_mol
                * model.fs.unit.cold_side.properties_in[0].cp_mol
            ),
            rel=1e-5,
        )
        assert value(model.fs.unit.NTU[0]) == pytest.approx(
            value(
                model.fs.unit.area * model.fs.unit.heat_transfer_coefficient[0] / Cmin
            ),
            rel=1e-5,
        )

        assert pytest.approx(0.7 * Cmin * (392.23 - 326.36), rel=1e-5) == value(
            model.fs.unit.heat_duty[0]
        )

        assert not model.fs.unit.hot_side_inlet.flow_mol[0].fixed
        assert not model.fs.unit.hot_side_inlet.temperature[0].fixed
        assert not model.fs.unit.hot_side_inlet.pressure[0].fixed
        assert not model.fs.unit.hot_side_inlet.mole_frac_comp[0, "CO2"].fixed
        assert not model.fs.unit.hot_side_inlet.mole_frac_comp[0, "H2O"].fixed
        assert not model.fs.unit.hot_side_inlet.mole_frac_comp[0, "MEA"].fixed

        assert not model.fs.unit.cold_side_inlet.flow_mol[0].fixed
        assert not model.fs.unit.cold_side_inlet.temperature[0].fixed
        assert not model.fs.unit.cold_side_inlet.pressure[0].fixed
        assert not model.fs.unit.cold_side_inlet.mole_frac_comp[0, "CO2"].fixed
        assert not model.fs.unit.cold_side_inlet.mole_frac_comp[0, "H2O"].fixed
        assert not model.fs.unit.cold_side_inlet.mole_frac_comp[0, "MEA"].fixed



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for costing base classes
"""
import pytest

from pyomo.environ import ConcreteModel, Constraint, Set, units as pyunits, Var, Param
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent

from idaes.core import declare_process_block_class, UnitModelBlockData
from idaes.core.util.model_statistics import degrees_of_freedom

from idaes.core import (
    FlowsheetCostingBlock,
    FlowsheetCostingBlockData,
    UnitModelCostingBlock,
    register_idaes_currency_units,
)

# TODO : Tests for cases with multiple costing packages
pyunits.load_definitions_from_strings(["USD_test = [test_currency]"])


@pytest.mark.unit
def test_register_idaes_currency_units():
    # Register standard currency units
    register_idaes_currency_units()

    assert "USD_CE500" in pyunits.pint_registry

    CEI = {
        "USD_1990": 357.6,
        "USD_1991": 361.3,
        "USD_1992": 358.2,
        "USD_1993": 359.2,
        "USD_1994": 368.1,
        "USD_1995": 381.1,
        "USD_1996": 381.7,
        "USD_1997": 386.5,
        "USD_1998": 389.5,
        "USD_1999": 390.6,
        "USD_2000": 394.1,
        "USD_2001": 394.3,
        "USD_2002": 395.6,
        "USD_2003": 402.0,
        "USD_2004": 444.2,
        "USD_2005": 468.2,
        "USD_2006": 499.6,
        "USD_2007": 525.4,
        "USD_2008": 575.4,
        "USD_2009": 521.9,
        "USD_2010": 550.8,
        "USD_2011": 585.7,
        "USD_2012": 584.6,
        "USD_2013": 567.3,
        "USD_2014": 576.1,
        "USD_2015": 556.8,
        "USD_2016": 541.7,
        "USD_2017": 567.5,
        "USD_2018": 603.1,
        "USD_2019": 607.5,
        "USD_2020": 596.2,
        "USD_2021": 708.0,
        "USD_2022": 816.0,
        "USD_2023": 797.9,
    }

    for c, conv in CEI.items():
        assert c in pyunits.pint_registry

        assert pytest.approx(conv / 500, rel=1e-10) == pyunits.convert_value(
            1, pyunits.USD_CE500, getattr(pyunits, c)
        )


class TestFlowsheetCostingBlock:
    @pytest.mark.unit
    def test_basic_attributes(self):
        m = ConcreteModel()

        with pytest.raises(
            ValueError,
            match="costing - costing package has not specified the base "
            "currency units to use for costing.",
        ):
            m.costing = FlowsheetCostingBlock()

        assert m.costing.base_currency is None
        assert m.costing.base_period is pyunits.year
        assert m.costing.defined_flows == {}
        assert FlowsheetCostingBlockData.unit_mapping == {}

    @pytest.mark.unit
    def test_build_global_params(self):
        with pytest.raises(
            NotImplementedError,
            match="Derived class has not defined a build_global_params method.",
        ):
            FlowsheetCostingBlockData.build_global_params(self)

    @pytest.mark.unit
    def test_build_process_costs(self):
        with pytest.raises(
            NotImplementedError,
            match="Derived class has not defined a build_process_costs method.",
        ):
            FlowsheetCostingBlockData.build_process_costs(self)

    @pytest.mark.unit
    def test_initialize_build(self):
        with pytest.raises(
            NotImplementedError,
            match="Derived class has not defined an initialize_build method.",
        ):
            FlowsheetCostingBlockData.initialize(self)


# Create some dummy classes to represent inherited unit models
@declare_process_block_class("TypeA")
class TypeAData(UnitModelBlockData):
    def build(self):
        super().build()
        self.class_type = "A"


@declare_process_block_class("TypeB")
class TypeBData(TypeAData):
    def build(self):
        super().build()
        self.class_type = "B"


@declare_process_block_class("TypeC")
class TypeCData(TypeBData):
    def build(self):
        super().build()
        self.class_type = "C"


@declare_process_block_class("TypeD")
class TypeDData(TypeAData):
    def build(self):
        super().build()
        self.class_type = "D"


@declare_process_block_class("TypeE")
class TypeEData(UnitModelBlockData):
    def build(self):
        super().build()
        self.class_type = "E"


@declare_process_block_class("DummyCostingPackage")
class DummyCostingPackageData(FlowsheetCostingBlockData):
    def build_global_params(self):
        self.base_currency = pyunits.USD_test
        self.base_period = pyunits.year

        self.test_flow_2_cost = Param(initialize=0.07, units=pyunits.kW)

        self.defined_flows = {
            "test_flow_1": 0.2 * pyunits.J,
            "test_flow_2": self.test_flow_2_cost,
        }

        self._bgp = True

    def build_process_costs(self):
        self._bpc = True

    def initialize_build(self):
        self._init = True

    def method_1(blk):
        blk.cost_method = 1

    def method_2(blk):
        blk.cost_method = 2

    def method_3(blk):
        blk.cost_method = 3

    def method_4(blk):
        blk.cost_method = 4

    unit_mapping = {TypeA: method_1, TypeB: method_2, TypeC: method_3}


class TestFlowsheetCostingBlock:
    @pytest.mark.unit
    def test_costing_package_no_base_currency(self):
        @declare_process_block_class("TestCostingPackage2")
        class TestCostingPackage2Data(DummyCostingPackageData):
            def build_global_params(self):
                super().build_global_params()
                self.base_currency = None

        m = ConcreteModel()

        with pytest.raises(
            ValueError,
            match="costing - costing package has not specified the base "
            "currency units to use for costing.",
        ):
            m.costing = TestCostingPackage2()

    @pytest.fixture(scope="class")
    def costing(self):
        m = ConcreteModel()
        m.costing = DummyCostingPackage()

        return m

    @pytest.mark.unit
    def test_basic_attributes(self, costing):
        assert costing.costing._registered_unit_costing == []
        assert isinstance(costing.costing.flow_types, Set)
        assert len(costing.costing.flow_types) == 2
        assert "test_flow_1" in costing.costing.flow_types
        assert costing.costing._registered_flows == {
            "test_flow_1": [],
            "test_flow_2": [],
        }

        assert costing.costing._costing_methods_map == {
            TypeAData: DummyCostingPackageData.method_1,
            TypeBData: DummyCostingPackageData.method_2,
            TypeCData: DummyCostingPackageData.method_3,
        }

        # Check that test_flow_1 was properly defined
        assert isinstance(costing.costing.test_flow_1_cost, Var)
        assert costing.costing.test_flow_1_cost.value == 0.2
        assert_units_equivalent(costing.costing.test_flow_1_cost.get_units(), pyunits.J)

        assert isinstance(costing.costing.test_flow_2_cost, Param)

        # Test that build_global_parameters was called successfully
        assert costing.costing._bgp

    @pytest.mark.unit
    def test_register_flow_type(self, costing):
        costing.costing.register_flow_type(
            "test_flow", 42 * pyunits.USD_test / pyunits.mol
        )

        assert isinstance(costing.costing.test_flow_cost, Var)
        assert costing.costing.test_flow_cost.value == 42
        assert_units_equivalent(
            costing.costing.test_flow_cost.get_units(), pyunits.USD_test / pyunits.mol
        )
        assert "test_flow" in costing.costing.flow_types

        assert costing.costing._registered_flows == {
            "test_flow_1": [],
            "test_flow_2": [],
            "test_flow": [],
        }

    @pytest.mark.unit
    def test_register_flow_component_exists(self, costing):
        costing.costing.test_flow_3_cost = Var()
        with pytest.raises(
            RuntimeError,
            match="Component test_flow_3_cost already exists on costing but is not 42.",
        ):
            costing.costing.register_flow_type("test_flow_3", 42)
        # cleanup for next test
        costing.costing.flow_types.remove("test_flow_3")
        costing.costing.del_component(costing.costing.test_flow_3_cost)

    @pytest.mark.unit
    def test_cost_flow_invalid_type(self, costing):
        with pytest.raises(
            ValueError,
            match="foo is not a recognized flow type. Please "
            "check your spelling and that the flow type has "
            "been registered with the FlowsheetCostingBlock.",
        ):
            costing.costing.cost_flow(42, "foo")

    @pytest.mark.unit
    def test_cost_flow_indexed_var(self, costing):
        costing.indexed_var = Var(
            [1, 2, 3], initialize=1, units=pyunits.mol / pyunits.s
        )
        with pytest.raises(
            TypeError,
            match="indexed_var is an indexed component. Flow "
            "costing only supports unindexed components.",
        ):
            costing.costing.cost_flow(costing.indexed_var, "test_flow")

    @pytest.mark.unit
    def test_cost_flow_unbounded_var(self, costing, caplog):
        costing.costing.cost_flow(costing.indexed_var[1], "test_flow")

        warn_str = (
            "indexed_var[1] has a lower bound of less "
            "than zero. Costing requires that all flows have a "
            "lower bound equal to or greater than zero to "
            "avoid negative costs."
        )

        assert warn_str in caplog.text

    @pytest.mark.unit
    def test_cost_flow_var(self, costing):
        costing.indexed_var[1].setlb(0)

        costing.costing.cost_flow(costing.indexed_var[1], "test_flow")

        assert costing.indexed_var[1] in costing.costing._registered_flows["test_flow"]

    @pytest.mark.unit
    def test_cost_flow_unbounded_expr(self, costing, caplog):
        costing.costing.cost_flow(-costing.indexed_var[2], "test_flow")

        warn_str = (
            "flow_expr is an expression with a lower "
            "bound of less than zero. Costing requires that "
            "all flows have a lower bound equal to or greater "
            "than zero to avoid negative costs."
        )

        assert warn_str in caplog.text

    @pytest.mark.unit
    def test_cost_flow_expr(self, costing):
        costing.indexed_var[2].setub(0)

        costing.costing.cost_flow(-costing.indexed_var[2], "test_flow")

        assert str(-costing.indexed_var[2]) == str(
            costing.costing._registered_flows["test_flow"][-1]
        )

    @pytest.mark.unit
    def test_get_costing_method_for(self, costing):
        costing.unit_a = TypeA()
        costing.unit_b = TypeB()
        costing.unit_c = TypeC()
        costing.unit_d = TypeD()
        costing.unit_e = TypeE()

        assert isinstance(costing.unit_a, TypeAData)

        assert (
            costing.costing._get_costing_method_for(costing.unit_a)
            is DummyCostingPackageData.method_1
        )

        assert (
            costing.costing._get_costing_method_for(costing.unit_b)
            is DummyCostingPackageData.method_2
        )

        assert (
            costing.costing._get_costing_method_for(costing.unit_c)
            is DummyCostingPackageData.method_3
        )

        # TypeD not registered with property package, but inherits from TypeA
        # Should get method_1
        assert (
            costing.costing._get_costing_method_for(costing.unit_d)
            is DummyCostingPackageData.method_1
        )

        # TypeE not registered with property package and no inheritance
        # Should return RuntimeError
        with pytest.raises(
            RuntimeError,
            match="Could not identify default costing method "
            "for unit_e. This implies the unit model's class "
            "and parent classes do not exist in the default "
            "mapping provided by the costing package. Please "
            "provide a specific costing method for this unit.",
        ):
            costing.costing._get_costing_method_for(costing.unit_e)

    @pytest.mark.unit
    def test_cost_unit_first(self, costing):
        costing.unit_a.costing = UnitModelCostingBlock(
            flowsheet_costing_block=costing.costing
        )

        assert costing.unit_a.costing.cost_method == 1

        assert costing.unit_a.costing in costing.unit_a._initialization_order
        assert costing.unit_a.costing in costing.costing._registered_unit_costing

    @pytest.mark.unit
    def test_del_unit_costing(self, costing):
        costing.unit_a.del_component(costing.unit_a.costing)

        assert not hasattr(costing.unit_a, "costing")
        assert costing.unit_a._initialization_order == []
        assert costing.costing._registered_unit_costing == []

    @pytest.mark.unit
    def test_cost_unit_duplicate(self, costing):
        costing.unit_a.costing = UnitModelCostingBlock(
            flowsheet_costing_block=costing.costing
        )

        # First, check implicit replacement
        # This should work
        costing.unit_a.costing = UnitModelCostingBlock(
            flowsheet_costing_block=costing.costing,
            costing_method=DummyCostingPackageData.method_2,
        )

        assert costing.unit_a.costing.cost_method == 2

        assert costing.unit_a.costing in costing.unit_a._initialization_order
        assert costing.unit_a.costing in costing.costing._registered_unit_costing

        # Then check double costing
        with pytest.raises(
            RuntimeError,
            match="Unit model unit_a already has a costing block "
            "registered: unit_a.costing. Each unit may only have a single "
            "UnitModelCostingBlock associated with it.",
        ):
            costing.unit_a.costing2 = UnitModelCostingBlock(
                flowsheet_costing_block=costing.costing
            )

        # Clean everything up at the end
        costing.unit_a.del_component(costing.unit_a.costing)
        costing.unit_a.del_component(costing.unit_a.costing2)

        # Make sure we cleaned up
        assert not hasattr(costing.unit_a, "costing")
        assert not hasattr(costing.unit_a, "costing2")

    @pytest.mark.unit
    def test_cost_unit_custom_method(self, costing):
        def custom_method(blk):
            blk.capital_cost = Var(
                initialize=1, bounds=(0, 1e10), units=pyunits.USD_test
            )
            blk.fixed_operating_cost = Var(
                initialize=1, bounds=(0, 1e10), units=pyunits.USD_test / pyunits.year
            )
            blk.variable_operating_cost = Var(
                initialize=1, bounds=(0, 1e10), units=pyunits.USD_test / pyunits.year
            )

            blk.capital_cost_constraint = Constraint(
                expr=blk.capital_cost == 4.2e6 * pyunits.USD_test
            )
            blk.fixed_operating_cost_constraint = Constraint(
                expr=blk.fixed_operating_cost == 1e2 * pyunits.USD_test / pyunits.year
            )
            blk.variable_operating_cost_constraint = Constraint(
                expr=blk.variable_operating_cost
                == 7e4 * pyunits.USD_test / pyunits.year
            )

            blk._checkvar = True

        costing.unit_a.costing = UnitModelCostingBlock(
            flowsheet_costing_block=costing.costing, costing_method=custom_method
        )

        assert isinstance(costing.unit_a.costing, UnitModelCostingBlock)
        assert costing.unit_a.costing in costing.unit_a._initialization_order
        assert costing.unit_a.costing in costing.costing._registered_unit_costing

        assert isinstance(costing.unit_a.costing.capital_cost, Var)
        assert isinstance(costing.unit_a.costing.variable_operating_cost, Var)
        assert isinstance(costing.unit_a.costing.fixed_operating_cost, Var)
        assert costing.unit_a.costing._checkvar

    @pytest.mark.unit
    def test_cost_unit_capital_cost_not_var(self, costing):
        def dummy_method(blk):
            blk.capital_cost = "foo"

        with pytest.raises(
            TypeError,
            match="unit_b capital_cost component must be a "
            "Var. Please check the costing package you are "
            "using to ensure that all costing components are "
            "declared as variables.",
        ):
            costing.unit_b.costing = UnitModelCostingBlock(
                flowsheet_costing_block=costing.costing, costing_method=dummy_method
            )

        # Clean up for next test
        costing.unit_b.del_component(costing.unit_b.costing)

    @pytest.mark.unit
    def test_cost_unit_capital_cost_lb(self, costing, caplog):
        def dummy_method(blk):
            blk.capital_cost = Var()

        costing.unit_b.costing = UnitModelCostingBlock(
            flowsheet_costing_block=costing.costing, costing_method=dummy_method
        )

        assert (
            "unit_b capital_cost component has a lower bound less than "
            "zero. Be aware that this may result in negative costs during "
            "optimization." in caplog.text
        )

        # Clean up for next test
        costing.unit_b.del_component(costing.unit_b.costing)

    @pytest.mark.unit
    def test_cost_unit_fixed_operating_cost_not_var(self, costing):
        def dummy_method(blk):
            blk.fixed_operating_cost = "foo"

        with pytest.raises(
            TypeError,
            match="unit_b fixed_operating_cost component must "
            "be a Var. Please check the costing package you "
            "are using to ensure that all costing components "
            "are declared as variables.",
        ):
            costing.unit_b.costing = UnitModelCostingBlock(
                flowsheet_costing_block=costing.costing, costing_method=dummy_method
            )

        # Clean up for next test
        costing.unit_b.del_component(costing.unit_b.costing)

    @pytest.mark.unit
    def test_cost_unit_fixed_operating_cost_lb(self, costing, caplog):
        def dummy_method(blk):
            blk.fixed_operating_cost = Var()

        costing.unit_b.costing = UnitModelCostingBlock(
            flowsheet_costing_block=costing.costing, costing_method=dummy_method
        )

        assert (
            "unit_b fixed_operating_cost component has a lower bound less "
            "than zero. Be aware that this may result in negative costs "
            "during optimization." in caplog.text
        )

        # Clean up for next test
        costing.unit_b.del_component(costing.unit_b.costing)

    @pytest.mark.unit
    def test_cost_unit_variable_operating_cost_not_var(self, costing):
        def dummy_method(blk):
            blk.variable_operating_cost = "foo"

        with pytest.raises(
            TypeError,
            match="unit_b variable_operating_cost component "
            "must be a Var. Please check the costing package "
            "you are using to ensure that all costing "
            "components are declared as variables.",
        ):
            costing.unit_b.costing = UnitModelCostingBlock(
                flowsheet_costing_block=costing.costing, costing_method=dummy_method
            )

        # Clean up for next test
        costing.unit_b.del_component(costing.unit_b.costing)

    @pytest.mark.unit
    def test_cost_unit_variable_operating_cost_lb(self, costing, caplog):
        def dummy_method(blk):
            blk.variable_operating_cost = Var()

        costing.unit_b.costing = UnitModelCostingBlock(
            flowsheet_costing_block=costing.costing, costing_method=dummy_method
        )

        assert (
            "unit_b variable_operating_cost component has a lower bound "
            "less than zero. Be aware that this may result in negative "
            "costs during optimization." in caplog.text
        )

        # Clean up for next test
        costing.unit_b.del_component(costing.unit_b.costing)

    @pytest.mark.unit
    def test_cost_process(self, costing):
        costing.costing.cost_process()

        # Check that build_process_costs was called from costing package
        assert costing.costing._bpc

        # Then check aggregation
        assert isinstance(costing.costing.aggregate_capital_cost, Var)
        assert str(costing.costing.aggregate_capital_cost.get_units()) == str(
            pyunits.USD_test
        )
        assert isinstance(costing.costing.aggregate_capital_cost_constraint, Constraint)

        assert isinstance(costing.costing.aggregate_fixed_operating_cost, Var)
        assert str(pyunits.USD_test / pyunits.year) == str(
            costing.costing.aggregate_fixed_operating_cost.get_units()
        )
        assert isinstance(
            costing.costing.aggregate_fixed_operating_cost_constraint, Constraint
        )

        assert isinstance(costing.costing.aggregate_variable_operating_cost, Var)
        assert str(pyunits.USD_test / pyunits.year) == str(
            costing.costing.aggregate_variable_operating_cost.get_units()
        )
        assert isinstance(
            costing.costing.aggregate_variable_operating_cost_constraint, Constraint
        )

        assert isinstance(costing.costing.aggregate_flow_test_flow, Var)
        assert str(pyunits.mol / pyunits.s) == str(
            costing.costing.aggregate_flow_test_flow.get_units()
        )
        assert isinstance(
            costing.costing.aggregate_flow_test_flow_constraint, Constraint
        )

        # We also have a test_flow_1 type registered, but no flows costed
        # This should have been skipped
        assert not hasattr(costing.costing, "aggregate_flow_test_flow_1")
        assert not hasattr(costing.costing, "aggregate_flow_test_flow_1_constraint")
        # unused flows do not get added to aggregate_flow_costs
        assert "test_flow_1" not in costing.costing.aggregate_flow_costs

        assert isinstance(costing.costing.aggregate_flow_costs, Var)
        assert str(pyunits.USD_test / pyunits.year) == str(
            costing.costing.aggregate_flow_costs.get_units()
        )
        assert len(costing.costing.aggregate_flow_costs) == 1
        assert isinstance(costing.costing.aggregate_flow_costs_constraint, Constraint)
        assert len(costing.costing.aggregate_flow_costs_constraint) == 1

    @pytest.mark.unit
    def test_unit_consistency(self, costing):
        assert_units_consistent(costing)

    @pytest.mark.unit
    def test_degrees_of_freedom(self, costing):
        costing.indexed_var[1].fix(2)
        costing.indexed_var[2].fix(-3)
        costing.indexed_var[3].fix(0)

        assert degrees_of_freedom(costing) == 0

    @pytest.mark.unit
    def test_initialize(self, costing):
        costing.costing.initialize()

        # Check that initialize was called from costing package
        assert costing.costing._init

        # Check that unit-level vars were initialized
        assert costing.unit_a.costing.capital_cost.value == 4.2e6
        assert costing.unit_a.costing.fixed_operating_cost.value == 100
        assert costing.unit_a.costing.variable_operating_cost.value == 7e4

        # Check that aggregate vars were initialized
        # Capital and operating costs should equal the unit level ones
        assert costing.costing.aggregate_capital_cost.value == 4.2e6
        assert costing.costing.aggregate_fixed_operating_cost.value == 100
        assert costing.costing.aggregate_variable_operating_cost.value == 7e4

        assert costing.costing.aggregate_flow_test_flow.value == 10

        assert pytest.approx(
            costing.costing.aggregate_flow_costs["test_flow"].value, rel=1e-12
        ) == (
            pyunits.convert_value(
                10 * 42, from_units=1 / pyunits.s, to_units=1 / pyunits.year
            )
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for data_utils module
"""
import pytest
import pandas as pd
from idaes.core.surrogate.sampling import (
    split_training_validation,
    split_training_validation_testing,
)


class TestDataUtils:
    @pytest.mark.unit
    def test_split_training_validation(self):
        d = {"a": [1, 2, 3, 4, 5], "b": [2, 3, 4, 5, 6]}
        df = pd.DataFrame(d)

        # test no seed
        df_training, df_validation = split_training_validation(
            dataframe=df, training_fraction=0.7
        )
        assert len(df_training) == 3
        assert len(df_validation) == 2

        # test seed
        expected_training_df = pd.DataFrame({"a": [2, 5, 3], "b": [3, 6, 4]})
        expected_validation_df = pd.DataFrame({"a": [1, 4], "b": [2, 5]})

        df_training, df_validation = split_training_validation(
            dataframe=df, training_fraction=0.7, seed=42
        )
        assert len(df_training) == 3
        assert len(df_validation) == 2
        pd.testing.assert_frame_equal(
            df_training.reset_index(drop=True), expected_training_df
        )
        pd.testing.assert_frame_equal(
            df_validation.reset_index(drop=True), expected_validation_df
        )

    @pytest.mark.unit
    def test_split_training_testing_validation(self):
        d = {"a": [1, 2, 3, 4, 5, 6, 7, 8, 9], "b": [2, 3, 4, 5, 6, 7, 8, 9, 10]}
        df = pd.DataFrame(d)

        # test no seed
        df_training, df_validation, df_testing = split_training_validation_testing(
            dataframe=df, training_fraction=0.6, validation_fraction=0.35
        )
        assert len(df_training) == 5
        assert len(df_validation) == 3
        assert len(df_testing) == 1

        # test seed
        expected_training_df = pd.DataFrame(
            {"a": [8, 2, 6, 1, 9], "b": [9, 3, 7, 2, 10]}
        )
        expected_validation_df = pd.DataFrame({"a": [3, 5, 4], "b": [4, 6, 5]})
        expected_testing_df = pd.DataFrame({"a": [7], "b": [8]})

        df_training, df_validation, df_testing = split_training_validation_testing(
            dataframe=df, training_fraction=0.6, validation_fraction=0.35, seed=42
        )

        assert len(df_training) == 5
        assert len(df_validation) == 3
        assert len(df_testing) == 1
        pd.testing.assert_frame_equal(
            df_training.reset_index(drop=True), expected_training_df
        )
        pd.testing.assert_frame_equal(
            df_validation.reset_index(drop=True), expected_validation_df
        )
        pd.testing.assert_frame_equal(
            df_testing.reset_index(drop=True), expected_testing_df
        )


if __name__ == "__main__":
    TestDataUtils().test_split_training_validation()
    TestDataUtils().test_split_training_validation_testing()
